import operator
import typing
import uuid

import sqlalchemy as sa
from sqlalchemy.exc import MultipleResultsFound, NoResultFound
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm.interfaces import ORMOption

T = typing.TypeVar("T")
_DT = typing.TypeVar("_DT")
_ChoiceLabelT = typing.TypeVar("_ChoiceLabelT")
_ChoiceValueT = typing.TypeVar("_ChoiceValueT")


class QueryError(Exception): ...


class NoResultError(QueryError, NoResultFound): ...


class MultipleResultsError(QueryError, MultipleResultsFound): ...


class ExecutableQuery(typing.Generic[T]):
    def __init__(self, dbsession: AsyncSession, stmt: sa.Select[tuple[T]]) -> None:
        self.dbsession = dbsession
        self.stmt = stmt

    async def one(self) -> T:
        """Return exactly one row or raise an exception."""
        try:
            rows = await self.dbsession.scalars(self.stmt)
            return rows.one()
        except NoResultFound as ex:
            raise NoResultError from ex
        except MultipleResultsFound as ex:
            raise MultipleResultsError from ex

    async def one_or_none(self) -> T | None:
        """Return exactly one row or None.
        Note, if there are more than one row, it will raise MultipleResultsError exception.

        :param stmt: SQL statement
        :raises MultipleResultsError: if more than one row is found
        :return: T | None
        """
        try:
            rows = await self.dbsession.scalars(self.stmt)
            return rows.one_or_none()
        except MultipleResultsFound as ex:
            raise MultipleResultsError from ex

    async def one_or_raise(self, exc: Exception) -> T:
        """Return exactly one row or raise a custom exception if no row exists."""
        entity = await self.one_or_none()
        if entity is None:
            raise exc
        return entity

    async def one_or_default(self, default_value: _DT) -> T | _DT:
        entity = await self.one_or_none()
        return entity if entity else default_value

    async def all(self) -> typing.Sequence[T]:
        """Return all rows as a collection."""
        result = await self.dbsession.scalars(self.stmt)
        return result.all()

    async def iterator(self, batch_size: int = 1000) -> typing.AsyncGenerator[T, None]:
        stmt = self.stmt.execution_options(yield_per=batch_size)
        result = await self.dbsession.stream(stmt)
        async for partition in result.partitions(batch_size):
            for row in partition:
                yield row[0]

    async def exists(self) -> bool:
        stmt = sa.select(sa.exists(self.stmt))
        result = await self.dbsession.scalars(stmt)
        return result.one() is True

    async def count(self) -> int:
        stmt = sa.select(sa.func.count()).select_from(self.stmt.subquery())
        result = await self.dbsession.scalars(stmt)
        count = result.one()
        return int(count) if count else 0

    @typing.overload
    async def choices(
        self,
        label_attr: str = "",
        value_attr: str = "",
    ) -> typing.AsyncGenerator[tuple[typing.Any, typing.Any], None]:  # pragma: no cover
        yield tuple(["", ""])  # type: ignore[misc]

    @typing.overload
    async def choices(
        self,
        label_attr: typing.Callable[[T], _ChoiceLabelT],
        value_attr: typing.Callable[[T], _ChoiceValueT],
    ) -> typing.AsyncGenerator[tuple[_ChoiceValueT, _ChoiceLabelT], None]:  # pragma: no cover
        yield tuple(["", ""])  # type: ignore[misc]

    async def choices(
        self,
        label_attr: typing.Any = str,
        value_attr: typing.Any = "id",
    ) -> typing.AsyncGenerator[tuple[_ChoiceValueT, _ChoiceLabelT], None]:
        label_getter: typing.Callable[[T], _ChoiceLabelT]
        value_getter: typing.Callable[[T], _ChoiceValueT]
        if isinstance(label_attr, str):
            label_getter = operator.attrgetter(label_attr)
        else:
            label_getter = label_attr

        if isinstance(value_attr, str):
            value_getter = operator.attrgetter(value_attr)
        else:
            value_getter = value_attr

        for item in await self.all():
            yield value_getter(item), label_getter(item)


class Query:
    def __init__(self, dbsession: AsyncSession) -> None:
        self.dbsession = dbsession

    @classmethod
    def of(cls, model_class: type[T]) -> sa.Select[tuple[T]]:
        """Return a select statement for the given model."""
        return sa.select(model_class)

    def select(self, stmt: sa.Select[tuple[T]]) -> ExecutableQuery[T]:
        return ExecutableQuery(self.dbsession, stmt)

    async def get(
        self, model_class: type[T], pk: int | str | uuid.UUID, *, options: typing.Sequence[ORMOption] = ()
    ) -> T | None:
        return await self.dbsession.get(model_class, pk, options=options)


query = Query



import functools
import typing

import anyio
import pytest
from starlette.requests import Request
from starlette.responses import PlainTextResponse, Response
from starlette.routing import Mount, Router
from starlette.testclient import TestClient
from starlette.websockets import WebSocket

from kupala.applications import Kupala
from kupala.dependencies import FactoryResolver
from kupala.responses import JSONResponse
from kupala.routing import (
    abs_url_for,
    AsyncViewCallable,
    CallNext,
    Route,
    route_matches,
    RouteGroup,
    url_matches,
    WebSocketRoute,
    WebSocketViewCallable,
)


@pytest.fixture
def route_group() -> RouteGroup:
    return RouteGroup()


async def example_set_middleware(request: Request, call_next: CallNext) -> Response:
    request.state.value = "set"
    return await call_next(request)


async def example_two_middleware(request: Request, call_next: CallNext) -> Response:
    request.state.value = request.state.value + "-two"
    return await call_next(request)


async def example_three_middleware(request: Request, call_next: CallNext) -> Response:
    request.state.value = request.state.value + "-three"
    return await call_next(request)


class TestRouteGroup:
    def test_getattr(self, route_group: RouteGroup) -> None:
        @route_group.get("/test")
        @route_group.get("/test/2")
        async def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        route = route_group[0]
        assert isinstance(route, Route)
        assert route.path == "/test/2"

        route = route_group[1]
        assert isinstance(route, Route)
        assert route.path == "/test"

        routes = route_group[:2]
        assert len(routes) == 2

    def test_repr(self, route_group: RouteGroup) -> None:
        @route_group.get("/test")
        async def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        assert repr(route_group) == "<RouteGroup: 1 route>"

    def test_iter(self, route_group: RouteGroup) -> None:
        @route_group.get("/test/{injection}")
        async def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        assert len(route_group) == 1

    def test_common_middleware(self) -> None:
        route_group = RouteGroup(middleware=[example_set_middleware])

        @route_group.get("/")
        async def view(request: Request) -> Response:
            return PlainTextResponse(request.state.value)

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.get("/")
            assert response.status_code == 200
            assert response.text == "set"

    def test_common_middleware_with_route_middleware(self) -> None:
        route_group = RouteGroup(middleware=[example_set_middleware])

        @route_group.get("/", middleware=[example_two_middleware])
        async def view(request: Request) -> Response:
            return PlainTextResponse(request.state.value)

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.get("/")
            assert response.status_code == 200
            assert response.text == "set-two"

    def test_app_middleware_only(self) -> None:
        route_group = RouteGroup(middleware=[])

        @route_group.get("/")
        async def view(request: Request) -> Response:
            return PlainTextResponse(request.state.value)

        app = Kupala(secret_key="key!", routes=route_group, middleware=[example_set_middleware])
        with TestClient(app) as client:
            response = client.get("/")
            assert response.status_code == 200
            assert response.text == "set"

    def test_app_route_group_middleware(self) -> None:
        route_group = RouteGroup(middleware=[example_two_middleware])

        @route_group.get("/")
        async def view(request: Request) -> Response:
            return PlainTextResponse(request.state.value)

        app = Kupala(secret_key="key!", routes=route_group, middleware=[example_set_middleware])
        with TestClient(app) as client:
            response = client.get("/")
            assert response.status_code == 200
            assert response.text == "set-two"

    def test_app_route_group_route_middleware(self) -> None:
        route_group = RouteGroup(middleware=[example_two_middleware])

        @route_group.get("/", middleware=[example_three_middleware])
        async def view(request: Request) -> Response:
            return PlainTextResponse(request.state.value)

        app = Kupala(secret_key="key!", routes=route_group, middleware=[example_set_middleware])
        with TestClient(app) as client:
            response = client.get("/")
            assert response.status_code == 200
            assert response.text == "set-two-three"

    def test_children(self) -> None:
        child_group = RouteGroup()

        @child_group.get("/test")
        def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        app = Kupala(secret_key="key!", routes=RouteGroup(children=[child_group]))
        with TestClient(app) as client:
            response = client.get("/test")
            assert response.status_code == 200
            assert response.text == "ok"

        assert client.post("/test").status_code == 405

    def test_children_with_route(self) -> None:
        def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        app = Kupala(
            secret_key="key!",
            routes=RouteGroup(
                children=[
                    Mount(path="/", routes=[Route("/test", view)]),
                ]
            ),
        )
        with TestClient(app) as client:
            response = client.get("/test")
            assert response.status_code == 200
            assert response.text == "ok"

            assert client.post("/test").status_code == 405

    def test_children_route_with_common_prefix(self) -> None:
        def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        app = Kupala(
            secret_key="key!",
            routes=RouteGroup(prefix="/api", children=[Route("/test", view)]),
        )
        with TestClient(app) as client:
            assert client.get("/api/test").status_code == 200
            assert client.post("/test").status_code == 404

    def test_children_route_group_with_common_prefix(self) -> None:
        def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        app = Kupala(
            secret_key="key!",
            routes=RouteGroup(
                prefix="/api",
                children=[
                    RouteGroup(
                        children=[Route("/test", view)],
                    ),
                ],
            ),
        )
        with TestClient(app) as client:
            assert client.get("/api/test").status_code == 200
            assert client.post("/test").status_code == 404

    def test_children_prefixed_route_group_with_common_prefix(self) -> None:
        def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        app = Kupala(
            secret_key="key!",
            routes=RouteGroup(
                prefix="/api",
                children=[
                    RouteGroup(
                        prefix="/v1",
                        children=[Route("/test", view)],
                    ),
                ],
            ),
        )
        with TestClient(app) as client:
            assert client.get("/api/v1/test").status_code == 200
            assert client.post("/api/test").status_code == 404
            assert client.post("/test").status_code == 404

    def test_get(self, route_group: RouteGroup) -> None:
        @route_group.get("/test")
        async def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.get("/test")
            assert response.status_code == 200
            assert response.text == "ok"

        assert client.post("/test").status_code == 405

    def test_get_named(self, route_group: RouteGroup) -> None:
        @route_group.get("/test", name="test")
        async def view(request: Request) -> Response:
            return PlainTextResponse(str(request.url_for("test")))

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.get("/test")
            assert response.text == "http://testserver/test"

    def test_get_middleware(self, route_group: RouteGroup) -> None:
        @route_group.get("/test", name="test", middleware=[example_set_middleware])
        async def view(request: Request) -> Response:
            return PlainTextResponse(request.scope["state"]["value"])

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.get("/test")
            assert response.text == "set"

    def test_get_or_post(self, route_group: RouteGroup) -> None:
        @route_group.get_or_post("/test")
        async def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.get("/test")
            assert response.status_code == 200
            assert response.text == "ok"

            response = client.post("/test")
            assert response.status_code == 200
            assert response.text == "ok"

        assert client.put("/test").status_code == 405

    def test_get_or_post_named(self, route_group: RouteGroup) -> None:
        @route_group.get_or_post("/test", name="test")
        async def view(request: Request) -> Response:
            return PlainTextResponse(str(request.url_for("test")))

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.post("/test")
            assert response.text == "http://testserver/test"

    def test_get_or_post_middleware(self, route_group: RouteGroup) -> None:
        @route_group.get_or_post("/test", name="test", middleware=[example_set_middleware])
        async def view(request: Request) -> Response:
            return PlainTextResponse(request.scope["state"]["value"])

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.post("/test")
            assert response.text == "set"

    def test_post_base(self, route_group: RouteGroup) -> None:
        @route_group.post("/test")
        async def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.post("/test")
            assert response.status_code == 200
            assert response.text == "ok"

        assert client.get("/test").status_code == 405

    def test_post_named(self, route_group: RouteGroup) -> None:
        @route_group.post("/test", name="test")
        async def view(request: Request) -> Response:
            return PlainTextResponse(str(request.url_for("test")))

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.post("/test")
            assert response.text == "http://testserver/test"

    def test_post_middleware(self, route_group: RouteGroup) -> None:
        @route_group.post("/test", name="test", middleware=[example_set_middleware])
        async def view(request: Request) -> Response:
            return PlainTextResponse(request.scope["state"]["value"])

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.post("/test")
            assert response.text == "set"

    def test_put_base(self, route_group: RouteGroup) -> None:
        @route_group.put("/test")
        async def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.put("/test")
            assert response.status_code == 200
            assert response.text == "ok"

        assert client.get("/test").status_code == 405

    def test_put_named(self, route_group: RouteGroup) -> None:
        @route_group.put("/test", name="test")
        async def view(request: Request) -> Response:
            return PlainTextResponse(str(request.url_for("test")))

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.put("/test")
            assert response.text == "http://testserver/test"

    def test_put_middleware(self, route_group: RouteGroup) -> None:
        @route_group.put("/test", name="test", middleware=[example_set_middleware])
        async def view(request: Request) -> Response:
            return PlainTextResponse(request.scope["state"]["value"])

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.put("/test")
            assert response.text == "set"

    def test_patch_base(self, route_group: RouteGroup) -> None:
        @route_group.patch("/test")
        async def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.patch("/test")
            assert response.status_code == 200
            assert response.text == "ok"

        assert client.get("/test").status_code == 405

    def test_patch_named(self, route_group: RouteGroup) -> None:
        @route_group.patch("/test", name="test")
        async def view(request: Request) -> Response:
            return PlainTextResponse(str(request.url_for("test")))

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.patch("/test")
            assert response.text == "http://testserver/test"

    def test_patch_middleware(self, route_group: RouteGroup) -> None:
        @route_group.patch("/test", name="test", middleware=[example_set_middleware])
        async def view(request: Request) -> Response:
            return PlainTextResponse(request.scope["state"]["value"])

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.patch("/test")
            assert response.text == "set"

    def test_delete_base(self, route_group: RouteGroup) -> None:
        @route_group.delete("/test")
        async def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.delete("/test")
            assert response.status_code == 200
            assert response.text == "ok"

        assert client.get("/test").status_code == 405

    def test_delete_named(self, route_group: RouteGroup) -> None:
        @route_group.delete("/test", name="test")
        async def view(request: Request) -> Response:
            return PlainTextResponse(str(request.url_for("test")))

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.delete("/test")
            assert response.text == "http://testserver/test"

    def test_delete_middleware(self, route_group: RouteGroup) -> None:
        @route_group.delete("/test", name="test", middleware=[example_set_middleware])
        async def view(request: Request) -> Response:
            return PlainTextResponse(request.scope["state"]["value"])

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            response = client.delete("/test")
            assert response.text == "set"

    def test_websocket_base(self, route_group: RouteGroup) -> None:
        @route_group.websocket("/test")
        async def view(websocket: WebSocket) -> None:
            await websocket.accept()
            await websocket.send_text("Hello, websocket!")
            await websocket.close()

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            with client.websocket_connect("/test") as session:
                assert session.receive_text() == "Hello, websocket!"

    def test_websocket_named_route(self, route_group: RouteGroup) -> None:
        @route_group.websocket("/test", name="test")
        async def view(websocket: WebSocket) -> None:
            await websocket.accept()
            await websocket.send_text(str(websocket.url_for("test")))
            await websocket.close()

        app = Kupala(secret_key="key!", routes=route_group)
        with TestClient(app) as client:
            with client.websocket_connect("/test") as session:
                assert session.receive_text() == "ws://testserver/test"


_Injection = typing.Annotated[str, "injected"]


class TestRoute:
    def test_injections(self) -> None:
        async def view(request: Request, injection: _Injection) -> Response:
            return PlainTextResponse(injection)

        app = Kupala(
            secret_key="key!",
            routes=[Route("/test/{injection}", view)],
        )
        with TestClient(app) as client:
            response = client.get("/test/injected")
            assert response.status_code == 200
            assert response.text == "injected"

    def test_injections_with_decorator(self) -> None:
        def view_decorator(fn: AsyncViewCallable) -> AsyncViewCallable:
            @functools.wraps(fn)
            async def inner_view(request: Request, **dependencies: typing.Any) -> Response:
                request.state.value = "fromdecorator"
                return await fn(request, **dependencies)

            return inner_view

        @view_decorator
        async def view(request: Request, injection: _Injection) -> Response:
            return PlainTextResponse(injection + request.state.value)

        app = Kupala(
            secret_key="key!",
            routes=[
                Route("/test/{injection}", view),
            ],
        )
        with TestClient(app) as client:
            response = client.get("/test/injected")
            assert response.status_code == 200
            assert response.text == "injectedfromdecorator"

    def test_injections_with_multiple_routes_on_same_view(self) -> None:
        async def view(request: Request, injection: _Injection) -> Response:
            return PlainTextResponse(injection)

        app = Kupala(
            secret_key="key!",
            routes=[
                Route("/test/{injection}", view),
                Route("/test/2/{injection}", view),
            ],
        )
        with TestClient(app) as client:
            assert client.get("/test/injected").text == "injected"
            assert client.get("/test/2/injected").text == "injected"

    def test_route_dependency_injects_request(self) -> None:
        def dependency(request: Request) -> str:
            return request.url.path

        RequestDep = typing.Annotated[str, FactoryResolver(dependency)]

        async def view(dep: RequestDep) -> Response:
            return PlainTextResponse(dep)

        app = Kupala(
            secret_key="key!",
            routes=[
                Route("/", view),
            ],
        )
        with TestClient(app) as client:
            response = client.get("/")
            assert response.status_code == 200
            assert response.text == "/"

    def test_norequest_handler(self) -> None:
        async def view() -> Response:
            return PlainTextResponse("ok")

        app = Kupala(
            secret_key="key!",
            routes=[
                Route("/test", view),
            ],
        )
        with TestClient(app) as client:
            response = client.get("/test")
            assert response.status_code == 200
            assert response.text == "ok"

        assert client.post("/test").status_code == 405

    def test_handler_with_decorator(self) -> None:
        def decorator(fn: AsyncViewCallable) -> AsyncViewCallable:
            @functools.wraps(fn)
            async def inner(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
                return await fn(*args, **kwargs)

            return inner

        @decorator
        async def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        app = Kupala(
            secret_key="key!",
            routes=[
                Route("/test", view),
            ],
        )
        with TestClient(app) as client:
            response = client.get("/test")
            assert response.status_code == 200
            assert response.text == "ok"

        assert client.post("/test").status_code == 405

    def test_async_handler(self) -> None:
        async def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        app = Kupala(
            secret_key="key!",
            routes=[
                Route("/test", view),
            ],
        )
        with TestClient(app) as client:
            response = client.get("/test")
            assert response.status_code == 200
            assert response.text == "ok"

        assert client.post("/test").status_code == 405

    def test_sync_handler(self) -> None:
        def view(request: Request) -> Response:
            return PlainTextResponse("ok")

        app = Kupala(
            secret_key="key!",
            routes=[
                Route("/test", view),
            ],
        )
        with TestClient(app) as client:
            response = client.get("/test")
            assert response.status_code == 200
            assert response.text == "ok"

        assert client.post("/test").status_code == 405

    def test_app_body_size_limit(self) -> None:
        async def view(request: Request) -> Response:
            return PlainTextResponse(await request.json())

        app = Kupala(
            secret_key="key!",
            routes=[
                Route("/test", view, methods=["POST"]),
            ],
            max_body_size=1,
        )
        with TestClient(app) as client:
            response = client.post("/test", json={"content": "a"})
            assert response.status_code == 413

    def test_route_body_size_limit(self) -> None:
        async def view(request: Request) -> Response:
            return JSONResponse(await request.json())

        app = Kupala(
            secret_key="key!",
            routes=[
                Route("/test", view, methods=["POST"], max_request_size=1),
            ],
        )
        with TestClient(app) as client:
            response = client.post("/test", json={"content": "a"})
            assert response.status_code == 413

    def test_route_increase_body_size_limit(self) -> None:
        async def view(request: Request) -> Response:
            return JSONResponse(await request.json())

        app = Kupala(
            secret_key="key!",
            max_body_size=1,
            routes=[
                Route("/test", view, methods=["POST"], max_request_size=1000000),
            ],
        )
        with TestClient(app) as client:
            response = client.post("/test", json={"content": "a"})
            assert response.status_code == 200

    def test_route_decrease_body_size_limit(self) -> None:
        async def view(request: Request) -> Response:
            return JSONResponse(await request.json())

        app = Kupala(
            secret_key="key!",
            max_body_size=1000000,
            routes=[
                Route("/test", view, methods=["POST"], max_request_size=1),
            ],
        )
        with TestClient(app) as client:
            response = client.post("/test", json={"content": "a"})
            assert response.status_code == 413

    def test_app_request_timeout(self) -> None:
        async def view(request: Request) -> Response:
            await anyio.sleep(0.002)
            return PlainTextResponse(await request.json())

        app = Kupala(
            secret_key="key!",
            routes=[
                Route("/test", view, methods=["POST"]),
            ],
            request_timeout=0.001,
        )
        with TestClient(app) as client:
            response = client.post("/test", json={"content": "a"})
            assert response.status_code == 504

    def test_route_request_timeout(self) -> None:
        async def view(request: Request) -> Response:
            await anyio.sleep(0.002)
            return JSONResponse(await request.json())

        app = Kupala(
            secret_key="key!",
            routes=[
                Route("/test", view, methods=["POST"], request_timeout=0.001),
            ],
        )
        with TestClient(app) as client:
            response = client.post("/test", json={"content": "a"})
            assert response.status_code == 504

    def test_route_increase_request_timeout(self) -> None:
        async def view(request: Request) -> Response:
            await anyio.sleep(0.002)
            return JSONResponse(await request.json())

        app = Kupala(
            secret_key="key!",
            request_timeout=0.001,
            routes=[
                Route("/test", view, methods=["POST"], request_timeout=0.005),
            ],
        )
        with TestClient(app) as client:
            response = client.post("/test", json={"content": "a"})
            assert response.status_code == 200

    def test_route_decrease_request_timeout(self) -> None:
        async def view(request: Request) -> Response:
            await anyio.sleep(0.002)
            return JSONResponse(await request.json())

        app = Kupala(
            secret_key="key!",
            request_timeout=0.1,
            routes=[
                Route("/test", view, methods=["POST"], request_timeout=0.001),
            ],
        )
        with TestClient(app) as client:
            response = client.post("/test", json={"content": "a"})
            assert response.status_code == 504


class TestWebSocketRoute:
    def test_websocket_dependency_injects_websocket(self) -> None:
        def dependency(websocket: WebSocket) -> str:
            return websocket.url.path

        RequestDep = typing.Annotated[str, FactoryResolver(dependency)]

        async def view(websocket: WebSocket, dep: RequestDep) -> None:
            await websocket.accept()
            await websocket.send_text(dep)
            await websocket.close()

        app = Kupala(
            secret_key="key!",
            routes=[
                WebSocketRoute("/test", view),
            ],
        )
        with TestClient(app) as client:
            with client.websocket_connect("/test") as session:
                assert session.receive_text() == "/test"

    def test_handler_with_decorator(self) -> None:
        def decorator(fn: WebSocketViewCallable) -> WebSocketViewCallable:
            @functools.wraps(fn)
            async def inner(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
                return await fn(*args, **kwargs)

            return inner

        @decorator
        async def view(websocket: WebSocket) -> None:
            await websocket.accept()
            await websocket.send_text("ok")
            await websocket.close()

        app = Kupala(
            secret_key="key!",
            routes=[
                WebSocketRoute("/test", view),
            ],
        )
        with TestClient(app) as client:
            with client.websocket_connect("/test") as session:
                assert session.receive_text() == "ok"


async def view(request: Request) -> Response:
    return PlainTextResponse("ok")


class TestURLHelpers:
    def test_url_matches(self) -> None:
        request = Request(
            scope={
                "type": "http",
                "method": "GET",
                "path": "/test",
                "headers": [(b"host", b"testserver")],
                "router": Router([Route("/test", view)]),
            }
        )
        assert url_matches(request, "/test")
        assert not url_matches(request, "/nottests")

    def test_route_matches(self) -> None:
        request = Request(
            scope={
                "type": "http",
                "method": "GET",
                "path": "/test",
                "headers": [(b"host", b"testserver")],
                "router": Router(
                    [
                        Route("/test", view, name="test"),
                        Route("/test/{param}", view, name="test-param"),
                    ]
                ),
            }
        )
        assert route_matches(request, "test")
        assert not route_matches(request, "test-param", {"param": "value"})

    def test_route_matches_with_params(self) -> None:
        request = Request(
            scope={
                "type": "http",
                "method": "GET",
                "path": "/test/value",
                "headers": [(b"host", b"testserver")],
                "router": Router(
                    [
                        Route("/test", view, name="test"),
                        Route("/test/{param}", view, name="test-param"),
                    ]
                ),
            }
        )
        assert route_matches(request, "test-param", {"param": "value"})
        assert not route_matches(request, "test-param", {"param": "value2"})

    def test_abs_url_for(self) -> None:
        request = Request(
            scope={
                "type": "http",
                "method": "GET",
                "path": "/test",
                "headers": [(b"host", b"testserver")],
                "router": Router([Route("/test", view, name="test")]),
            }
        )
        assert abs_url_for(request, "test") == "http://testserver/test"



from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession
from starlette.responses import Response
from starlette.testclient import TestClient

from kupala.applications import Kupala
from kupala.database.dependencies import DbSession
from kupala.database.extension import SQLAlchemy
from kupala.database.manager import Database
from kupala.routing import Route


async def inject_async_session_view(dbsession: AsyncSession) -> Response:
    return Response(dbsession.__class__.__name__)


async def inject_db_session_view(dbsession: DbSession) -> Response:
    return Response(dbsession.__class__.__name__)


def test_provides_async_session_dependency() -> None:
    app = Kupala(
        secret_key="key!",
        routes=[
            Route("/", inject_async_session_view),
        ],
        extensions=[
            SQLAlchemy(
                {
                    "default": Database("sqlite+aiosqlite:///:memory:"),
                }
            )
        ],
    )
    with TestClient(app) as client:
        response = client.get("/")
        assert response.status_code == 200
        assert response.text == "AsyncSession"


def test_provides_dbsession_dependency() -> None:
    app = Kupala(
        secret_key="key!",
        routes=[
            Route("/", inject_db_session_view),
        ],
        extensions=[
            SQLAlchemy(
                {
                    "default": Database("sqlite+aiosqlite:///:memory:"),
                }
            )
        ],
    )
    with TestClient(app) as client:
        response = client.get("/")
        assert response.status_code == 200
        assert response.text == "AsyncSession"


async def test_get_database() -> None:
    app = Kupala(
        secret_key="key!",
        routes=[
            Route("/", inject_db_session_view),
        ],
        extensions=[
            SQLAlchemy(
                {
                    "default": Database("sqlite+aiosqlite:///:memory:"),
                    "secondary": Database("sqlite+aiosqlite:///:memory:"),
                }
            )
        ],
    )
    async with app.initialize():
        database = SQLAlchemy.of(app).get_database("default")
        assert isinstance(database, Database)

        database = SQLAlchemy.of(app).get_database("secondary")
        assert isinstance(database, Database)


async def test_new_session() -> None:
    app = Kupala(
        secret_key="key!",
        routes=[
            Route("/", inject_db_session_view),
        ],
        extensions=[
            SQLAlchemy(
                {
                    "default": Database("sqlite+aiosqlite:///:memory:"),
                    "secondary": Database("sqlite+aiosqlite:///tmp/db"),
                }
            )
        ],
    )
    async with app.initialize():
        db = SQLAlchemy.of(app)
        async with db.new_session("default") as dbsession:
            assert isinstance(dbsession, AsyncSession)
            assert isinstance(dbsession.bind, AsyncEngine)
            assert str(dbsession.bind.url) == "sqlite+aiosqlite:///:memory:"

        async with db.new_session("secondary") as dbsession:
            assert isinstance(dbsession, AsyncSession)
            assert isinstance(dbsession.bind, AsyncEngine)
            assert str(dbsession.bind.url) == "sqlite+aiosqlite:///tmp/db"



import pytest

from kupala.passwords import PasswordHasher


@pytest.fixture
def passwords() -> PasswordHasher:
    return PasswordHasher()


class TestPasswords:
    def test_password_verification(self, passwords: PasswordHasher) -> None:
        plain_password = "password"
        hashed_password = passwords.make(plain_password)
        assert passwords.verify(plain_password, hashed_password)

    async def test_password_async_verification(self, passwords: PasswordHasher) -> None:
        plain_password = "password"
        hashed_password = await passwords.amake(plain_password)
        assert await passwords.averify(plain_password, hashed_password)

    def test_password_migration(self) -> None:
        plain_password = "password"

        passwords_sha256 = PasswordHasher(["pbkdf2_sha256"])
        passwords_sha512 = PasswordHasher(["pbkdf2_sha512"])
        hashed_password_sha512 = passwords_sha512.make(plain_password)
        hashed_password_sha256 = passwords_sha256.make(plain_password)

        passwords = PasswordHasher(["pbkdf2_sha256", "pbkdf2_sha512"])
        assert passwords.verify(plain_password, hashed_password_sha512)
        assert passwords.verify(plain_password, hashed_password_sha256)

    def test_needs_update(self) -> None:
        plain_password = "password"
        old_passwords = PasswordHasher(["pbkdf2_sha256"])
        new_passwords = PasswordHasher(
            ["pbkdf2_sha512", "pbkdf2_sha256"],
            deprecated=["pbkdf2_sha256"],
            default="pbkdf2_sha512",
        )

        password = old_passwords.make(plain_password)
        assert new_passwords.needs_update(password)

    async def test_verify_and_migrate(self) -> None:
        plain_password = "password"
        old_passwords = PasswordHasher(["pbkdf2_sha256"])
        new_passwords = PasswordHasher(
            ["pbkdf2_sha512", "pbkdf2_sha256"],
            deprecated=["pbkdf2_sha256"],
            default="pbkdf2_sha512",
        )

        password = old_passwords.make(plain_password)
        migrated, new_hash = new_passwords.verify_and_migrate(plain_password, password)
        assert migrated
        assert new_hash
        assert new_passwords.verify(plain_password, new_hash)

        migrated, new_hash = await new_passwords.averify_and_migrate(plain_password, password)
        assert migrated
        assert new_hash
        assert new_passwords.verify(plain_password, new_hash)



from __future__ import annotations

import contextlib

from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.database.manager import Database


class DbSessionMiddleware:
    def __init__(self, app: ASGIApp, databases: dict[str, Database]) -> None:
        self.app = app
        self.databases = databases

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        scope.setdefault("state", {})
        async with contextlib.AsyncExitStack() as stack:
            for name, database in self.databases.items():
                dbsession = await stack.enter_async_context(database.session())
                await stack.enter_async_context(dbsession)
                if name == "default":
                    scope["state"]["dbsession"] = dbsession
                scope["state"][f"dbsession_{name}"] = dbsession

            await self.app(scope, receive, send)



import contextlib
import inspect
import typing

import pytest
from starlette.requests import HTTPConnection, Request

from kupala.dependencies import (
    Argument,
    create_dependency_from_parameter,
    create_dependency_specs,
    Dependencies,
    DependencyError,
    DependencyNotFoundError,
    FactoryResolver,
    FromPath,
    RegistryResolver,
    RequestResolver,
    ResolveContext,
    VariableResolver,
)


class TestVariableResolver:
    async def test_variable_resolver(self) -> None:
        resolver = VariableResolver("abc")
        context = ResolveContext(
            sync_stack=contextlib.ExitStack(),
            async_stack=contextlib.AsyncExitStack(),
            dependencies=Dependencies(),
        )
        value = await resolver.resolve(context)
        assert value == "abc"


class TestFactoryResolver:
    async def test_sync_factory(self) -> None:
        def factory() -> str:
            return "abc"

        resolver = FactoryResolver(factory)
        context = ResolveContext(
            sync_stack=contextlib.ExitStack(),
            async_stack=contextlib.AsyncExitStack(),
            dependencies=Dependencies(
                {
                    Argument: VariableResolver(
                        Argument(param_name="arg", param_type=str, default=None, annotation=str, nullable=False)
                    ),
                }
            ),
        )
        value = await resolver.resolve(context)
        assert value == "abc"

    async def test_async_factory(self) -> None:
        async def factory() -> str:
            return "abc"

        resolver = FactoryResolver(factory)
        context = ResolveContext(
            sync_stack=contextlib.ExitStack(),
            async_stack=contextlib.AsyncExitStack(),
            dependencies=Dependencies(
                {
                    Argument: VariableResolver(
                        Argument(param_name="arg", param_type=str, default=None, annotation=str, nullable=False)
                    ),
                }
            ),
        )
        value = await resolver.resolve(context)
        assert value == "abc"

    async def test_subdependencies(self) -> None:
        def level3_factory() -> str:
            return "three"

        def level2_factory(three: typing.Annotated[str, FactoryResolver(level3_factory)]) -> str:
            return f"two-{three}"

        def level1_factory(two: typing.Annotated[str, FactoryResolver(level2_factory)]) -> str:
            return f"one-{two}"

        def factory(parent: typing.Annotated[str, FactoryResolver(level1_factory)]) -> str:
            return f"ok-{parent}"

        resolver = FactoryResolver(factory)
        context = ResolveContext(
            sync_stack=contextlib.ExitStack(),
            async_stack=contextlib.AsyncExitStack(),
            dependencies=Dependencies(
                {
                    Argument: VariableResolver(
                        Argument(param_name="arg", param_type=str, default=None, annotation=str, nullable=False)
                    ),
                }
            ),
        )
        value = await resolver.resolve(context)
        assert value == "ok-one-two-three"

    async def test_async_subdependencies(self) -> None:
        async def level3_factory() -> str:
            return "three"

        async def level2_factory(three: typing.Annotated[str, FactoryResolver(level3_factory)]) -> str:
            return f"two-{three}"

        async def level1_factory(two: typing.Annotated[str, FactoryResolver(level2_factory)]) -> str:
            return f"one-{two}"

        async def factory(parent: typing.Annotated[str, FactoryResolver(level1_factory)]) -> str:
            return f"ok-{parent}"

        resolver = FactoryResolver(factory)
        context = ResolveContext(
            sync_stack=contextlib.ExitStack(),
            async_stack=contextlib.AsyncExitStack(),
            dependencies=Dependencies(
                {
                    Argument: VariableResolver(
                        Argument(param_name="arg", param_type=str, default=None, annotation=str, nullable=False)
                    ),
                }
            ),
        )
        value = await resolver.resolve(context)
        assert value == "ok-one-two-three"

    async def test_mixed_subdependencies(self) -> None:
        async def level3_factory() -> str:
            return "three"

        def level2_factory(three: typing.Annotated[str, FactoryResolver(level3_factory)]) -> str:
            return f"two-{three}"

        async def level1_factory(two: typing.Annotated[str, FactoryResolver(level2_factory)]) -> str:
            return f"one-{two}"

        def factory(parent: typing.Annotated[str, FactoryResolver(level1_factory)]) -> str:
            return f"ok-{parent}"

        resolver = FactoryResolver(factory)
        context = ResolveContext(
            sync_stack=contextlib.ExitStack(),
            async_stack=contextlib.AsyncExitStack(),
            dependencies=Dependencies(
                {
                    Argument: VariableResolver(
                        Argument(param_name="arg", param_type=str, default=None, annotation=str, nullable=False)
                    ),
                }
            ),
        )
        value = await resolver.resolve(context)
        assert value == "ok-one-two-three"

    async def test_sync_context_manager_dependency(self) -> None:
        @contextlib.contextmanager
        def factory() -> typing.Generator[str, None, None]:
            yield "abc"

        resolver = FactoryResolver(factory)
        context = ResolveContext(
            sync_stack=contextlib.ExitStack(),
            async_stack=contextlib.AsyncExitStack(),
            dependencies=Dependencies(
                {
                    Argument: VariableResolver(
                        Argument(param_name="arg", param_type=str, default=None, annotation=str, nullable=False)
                    ),
                }
            ),
        )
        async with context:
            value = await resolver.resolve(context)
        assert value == "abc"

    async def test_async_context_manager_dependency(self) -> None:
        @contextlib.asynccontextmanager
        async def factory() -> typing.AsyncGenerator[str, None]:
            yield "abc"

        resolver = FactoryResolver(factory)
        context = ResolveContext(
            sync_stack=contextlib.ExitStack(),
            async_stack=contextlib.AsyncExitStack(),
            dependencies=Dependencies(
                {
                    Argument: VariableResolver(
                        Argument(param_name="arg", param_type=str, default=None, annotation=str, nullable=False)
                    ),
                }
            ),
        )
        async with context:
            value = await resolver.resolve(context)
        assert value == "abc"


class TestFactoryResolverScopes:
    async def test_transient(self) -> None:
        def factory() -> object:
            return object()

        resolver = FactoryResolver(factory, scope="transient")
        context = ResolveContext(
            sync_stack=contextlib.ExitStack(),
            async_stack=contextlib.AsyncExitStack(),
            dependencies=Dependencies(
                {
                    Argument: VariableResolver(
                        Argument(param_name="arg", param_type=str, default=None, annotation=str, nullable=False)
                    ),
                }
            ),
        )
        value1 = await resolver.resolve(context)
        value2 = await resolver.resolve(context)
        assert id(value1) != id(value2)

    async def test_cached(self) -> None:
        def factory() -> object:
            return object()

        resolver = FactoryResolver(factory, scope="cached")
        context = ResolveContext(
            sync_stack=contextlib.ExitStack(),
            async_stack=contextlib.AsyncExitStack(),
            dependencies=Dependencies(
                {
                    Argument: VariableResolver(
                        Argument(param_name="arg", param_type=str, default=None, annotation=str, nullable=False)
                    ),
                }
            ),
        )
        value1 = await resolver.resolve(context)
        value2 = await resolver.resolve(context)
        assert id(value1) == id(value2)

    async def test_scoped(self) -> None:
        def factory() -> object:
            return object()

        resolver = FactoryResolver(factory, scope="scoped")
        context = ResolveContext(
            sync_stack=contextlib.ExitStack(),
            async_stack=contextlib.AsyncExitStack(),
            dependencies=Dependencies(
                {
                    Argument: VariableResolver(
                        Argument(param_name="arg", param_type=str, default=None, annotation=str, nullable=False)
                    ),
                }
            ),
        )
        context2 = ResolveContext(
            sync_stack=contextlib.ExitStack(),
            async_stack=contextlib.AsyncExitStack(),
            dependencies=Dependencies(
                {
                    Argument: VariableResolver(
                        Argument(param_name="arg", param_type=str, default=None, annotation=str, nullable=False)
                    ),
                }
            ),
        )
        value1: object = await resolver.resolve(context)
        value2: object = await resolver.resolve(context)
        value3: object = await resolver.resolve(context2)
        assert id(value1) == id(value2)
        assert id(value1) != id(value3)
        assert id(value2) != id(value3)


class TestRequestResolver:
    async def test_one_param_callback(self) -> None:
        request = Request({"type": "http", "state": {"dep": "abc"}})

        resolver = RequestResolver(lambda r: r.state.dep)
        context = ResolveContext(
            sync_stack=contextlib.ExitStack(),
            async_stack=contextlib.AsyncExitStack(),
            dependencies=Dependencies(
                {
                    HTTPConnection: VariableResolver(request),
                }
            ),
        )
        assert await resolver.resolve(context) == "abc"

    async def test_two_param_callback(self) -> None:
        request = Request({"type": "http", "state": {"dep": "abc"}})

        resolver = RequestResolver(lambda r, a: r.state.dep + a.param_name)
        context = ResolveContext(
            sync_stack=contextlib.ExitStack(),
            async_stack=contextlib.AsyncExitStack(),
            dependencies=Dependencies(
                {
                    HTTPConnection: VariableResolver(request),
                    Argument: VariableResolver(
                        Argument(
                            param_name="def",
                            param_type=str,
                            default=None,
                            annotation=str,
                            nullable=False,
                        )
                    ),
                }
            ),
        )
        assert await resolver.resolve(context) == "abcdef"


class TestRegistryResolver:
    async def test_resolve(self) -> None:
        resolver = RegistryResolver(Dep)
        context = ResolveContext(
            sync_stack=contextlib.ExitStack(),
            async_stack=contextlib.AsyncExitStack(),
            dependencies=Dependencies(
                {
                    Dep: VariableResolver("abc"),
                }
            ),
        )
        value: str = await resolver.resolve(context)
        assert value == "abc"


class TestDependencies:
    async def test_resolves_dependency(self) -> None:
        dependencies = Dependencies(
            dependencies={
                str: VariableResolver("abc"),
            }
        )
        context = ResolveContext(dependencies=Dependencies())
        assert await dependencies.resolve(context, str) == "abc"

    async def test_resolves_from_parent(self) -> None:
        dependencies = Dependencies(
            dependencies={str: VariableResolver("abc")},
            parent=Dependencies(dependencies={int: VariableResolver(1)}),
        )

        context = ResolveContext(dependencies=Dependencies())
        assert await dependencies.resolve(context, int) == 1

    async def test_resolves_type_alias(self) -> None:
        type A = str
        dependencies = Dependencies(
            dependencies={
                A: VariableResolver("abc"),
                str: VariableResolver("str"),
            }
        )

        context = ResolveContext(dependencies=Dependencies())
        assert await dependencies.resolve(context, A) == "abc"
        assert await dependencies.resolve(context, str) == "str"

    async def test_missing_type(self) -> None:
        context = ResolveContext(dependencies=Dependencies())
        with pytest.raises(DependencyNotFoundError):
            await context.resolve(str)

    async def test_call_with(self) -> None:
        type Dep = str

        def fn(dep: Dep, dep2: typing.Annotated[str, VariableResolver("def")]) -> str:
            return dep + dep2

        resolver = Dependencies(
            dependencies={
                Dep: VariableResolver("abc"),
            }
        )
        dependencies = create_dependency_specs(fn)
        assert await resolver.call_with(fn, dependencies) == "abcdef"

    async def test_call_with_async(self) -> None:
        type Dep = str

        async def fn(dep: Dep, dep2: typing.Annotated[str, VariableResolver("def")]) -> str:
            return dep + dep2

        resolver = Dependencies(
            dependencies={
                Dep: VariableResolver("abc"),
            }
        )
        dependencies = create_dependency_specs(fn)
        assert await resolver.call_with(fn, dependencies) == "abcdef"  # type: ignore


type Dep = str
type DepAlias = str


class TestCreateDependencyFromParameter:
    async def test_type(self) -> None:
        """It should create an injection from a simple annotation.
        Example def fn(dep: Dep) -> None: ...
        """
        injection = create_dependency_from_parameter(
            inspect.Parameter(
                name="dep",
                annotation=Dep,
                default=inspect.Parameter.empty,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )
        assert injection.argument.param_name == "dep"
        assert injection.argument.param_type == Dep
        assert injection.argument.default is inspect.Parameter.empty
        assert injection.argument.annotation == Dep
        assert injection.argument.nullable is False

    def test_union_unsupported(self) -> None:
        """It should raise an error if the union has more than two parameters.
        Example def fn(dep: str | int | None) -> None: ..."""
        with pytest.raises(DependencyError, match="Union type with more than two arguments is not supported"):
            create_dependency_from_parameter(
                inspect.Parameter(
                    name="dep",
                    default=inspect.Parameter.empty,
                    annotation=str | int | None,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                )
            )

    def test_param_default_unsupported(self) -> None:
        with pytest.raises(DependencyError, match="Dependencies with default values are not supported."):
            create_dependency_from_parameter(
                inspect.Parameter(
                    name="dep",
                    default="abc",
                    annotation=str,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                )
            )

    def test_positional_only_unsupported(self) -> None:
        with pytest.raises(DependencyError, match="Dependencies with positional-only parameters are not supported"):
            create_dependency_from_parameter(
                inspect.Parameter(
                    name="dep",
                    default="abc",
                    annotation=str,
                    kind=inspect.Parameter.POSITIONAL_ONLY,
                )
            )

    def test_union_none(self) -> None:
        """It should create an injection from a nullable type.
        Example def fn(dep: str | None) -> None: ..."""
        injection = create_dependency_from_parameter(
            inspect.Parameter(
                name="dep",
                default=inspect.Parameter.empty,
                annotation=str | None,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )
        assert injection.argument.param_type is str
        assert injection.argument.nullable is True

    def test_union_non_none(self) -> None:
        """It should raise an error if the union has no None type.
        Example def fn(dep: str | int) -> None: ..."""
        with pytest.raises(DependencyError, match="Union type without None is not supported"):
            create_dependency_from_parameter(
                inspect.Parameter(
                    name="dep",
                    default=inspect.Parameter.empty,
                    annotation=str | int,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                )
            )

    def test_annotated(self) -> None:
        """It should create an injection from an annotated type.
        Example def fn(dep: typing.Annotated[Dep, VariableResolver("abc")]) -> None: ...
        """
        injection = create_dependency_from_parameter(
            inspect.Parameter(
                name="dep",
                default=inspect.Parameter.empty,
                annotation=typing.Annotated[str, VariableResolver("abc")],
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )
        assert injection.argument.param_type is str
        assert injection.argument.nullable is False
        assert isinstance(injection.resolver, VariableResolver)

    def test_annotated_optional(self) -> None:
        """It should create an injection from an annotated nullable type.
        Example def fn(dep: Dep | None) -> None: ..."""
        with pytest.raises(DependencyError, match="Only Union and Annotated"):
            create_dependency_from_parameter(
                inspect.Parameter(
                    name="dep",
                    default=inspect.Parameter.empty,
                    annotation=typing.Annotated[str, VariableResolver("abc")] | None,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                )
            )

    def test_annotated_union(self) -> None:
        """It should create an injection from an annotated union type.
        Example def fn(dep: typing.Annotated[Dep, VariableResolver("abc")] | int) -> None: ...
        """
        with pytest.raises(DependencyError, match="not supported"):
            create_dependency_from_parameter(
                inspect.Parameter(
                    name="dep",
                    default=inspect.Parameter.empty,
                    annotation=typing.Annotated[str | int, VariableResolver("abc")],
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                )
            )

    def test_annotated_nullable(self) -> None:
        injection = create_dependency_from_parameter(
            inspect.Parameter(
                name="dep",
                default=inspect.Parameter.empty,
                annotation=typing.Annotated[str | None, VariableResolver("abc")],
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )
        assert injection.argument.param_type is str
        assert injection.argument.nullable is True

    def test_annotated_typevar_with_default_type(self) -> None:
        _T = typing.TypeVar("_T", default=int)
        _Dep = typing.Annotated[_T, VariableResolver("abc")]

        injection = create_dependency_from_parameter(
            inspect.Parameter(
                name="dep",
                default=inspect.Parameter.empty,
                annotation=_Dep,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )
        assert injection.argument.param_type is int

    def test_annotated_typevar_without_default_type(self) -> None:
        _T = typing.TypeVar("_T")
        _Dep = typing.Annotated[_T, VariableResolver("abc")]

        with pytest.raises(DependencyError, match="TypeVar without default type is not supported"):
            create_dependency_from_parameter(
                inspect.Parameter(
                    name="dep",
                    default=inspect.Parameter.empty,
                    annotation=_Dep,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                )
            )

    def test_annotated_resolver(self) -> None:
        injection = create_dependency_from_parameter(
            inspect.Parameter(
                name="dep",
                default=inspect.Parameter.empty,
                annotation=typing.Annotated[str, VariableResolver("abc")],
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )
        assert injection.argument.param_type is str
        assert isinstance(injection.resolver, VariableResolver)

    def test_annotated_resolver_with_extra_annotations(self) -> None:
        injection = create_dependency_from_parameter(
            inspect.Parameter(
                name="dep",
                default=inspect.Parameter.empty,
                annotation=typing.Annotated[str, "abc", VariableResolver("abc")],
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )
        assert injection.argument.param_type is str
        assert isinstance(injection.resolver, VariableResolver)

    async def test_annotated_generic(self) -> None:
        class MyModel: ...

        async def _from_json_resolver(arg: Argument) -> object:
            return arg.param_type

        S = typing.TypeVar("S")
        FromJSON = typing.Annotated[S, FactoryResolver(_from_json_resolver)]

        injection = create_dependency_from_parameter(
            inspect.Parameter(
                name="dep",
                annotation=FromJSON[MyModel],
                default=inspect.Parameter.empty,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )
        context = ResolveContext(
            dependencies=Dependencies(
                {
                    HTTPConnection: VariableResolver(Request({"type": "http"})),
                    Argument: VariableResolver(injection.argument),
                }
            ),
        )
        assert injection.argument.param_type is MyModel
        assert await context.resolve(injection.resolver) == MyModel

    def test_unsupported_generic(self) -> None:
        with pytest.raises(DependencyError, match="are supported as dependencies"):
            create_dependency_from_parameter(
                inspect.Parameter(
                    name="dep",
                    default=inspect.Parameter.empty,
                    annotation=list[str],
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                )
            )

    async def test_shortcut_variable(self) -> None:
        injection = create_dependency_from_parameter(
            inspect.Parameter(
                name="dep",
                annotation=typing.Annotated[str, "value"],
                default=inspect.Parameter.empty,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )
        context = ResolveContext(
            dependencies=Dependencies({HTTPConnection: VariableResolver(Request({"type": "http"}))}),
        )
        assert injection.argument.param_type is str
        assert await context.resolve(injection.resolver) == "value"

    async def test_shortcut_factory(self) -> None:
        def factory() -> str:
            return "value"

        injection = create_dependency_from_parameter(
            inspect.Parameter(
                name="dep",
                annotation=typing.Annotated[str, factory],
                default=inspect.Parameter.empty,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )
        context = ResolveContext(
            dependencies=Dependencies(
                {
                    HTTPConnection: VariableResolver(Request({"type": "http"})),
                    Argument: VariableResolver(
                        Argument(param_name="arg", param_type=str, default=None, annotation=str, nullable=False)
                    ),
                }
            ),
        )
        assert await context.resolve(injection.resolver) == "value"

    async def test_shortcut_zero_arg_lambda(self) -> None:
        injection = create_dependency_from_parameter(
            inspect.Parameter(
                name="dep",
                annotation=typing.Annotated[str, lambda: "value"],
                default=inspect.Parameter.empty,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )

        context = ResolveContext(
            dependencies=Dependencies(
                {
                    HTTPConnection: VariableResolver(Request({"type": "http"})),
                    Argument: VariableResolver(
                        Argument(param_name="arg", param_type=str, default=None, annotation=str, nullable=False)
                    ),
                }
            ),
        )
        assert await context.resolve(injection.resolver) == "value"

    async def test_shortcut_one_arg_lambda(self) -> None:
        injection = create_dependency_from_parameter(
            inspect.Parameter(
                name="dep",
                annotation=typing.Annotated[str, lambda r: "value"],
                default=inspect.Parameter.empty,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )

        context = ResolveContext(
            dependencies=Dependencies({HTTPConnection: VariableResolver(Request({"type": "http"}))}),
        )
        assert await context.resolve(injection.resolver) == "value"

    async def test_shortcut_two_arg_lambda(self) -> None:
        injection = create_dependency_from_parameter(
            inspect.Parameter(
                name="dep",
                annotation=typing.Annotated[str, lambda r, arg: "value" + arg.param_name],
                default=inspect.Parameter.empty,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )

        context = ResolveContext(
            dependencies=Dependencies(
                {
                    HTTPConnection: VariableResolver(Request({"type": "http"})),
                    Argument: VariableResolver(
                        Argument(
                            param_name="arg",
                            param_type=str,
                            default=None,
                            annotation=str,
                            nullable=False,
                        )
                    ),
                }
            ),
        )
        assert await context.resolve(injection.resolver) == "valuearg"

    def test_shortcut_three_arg_lambda(self) -> None:
        with pytest.raises(DependencyError, match="should accept zero, one, or two parameters"):
            create_dependency_from_parameter(
                inspect.Parameter(
                    name="dep",
                    annotation=typing.Annotated[str, lambda r, arg, bar: "value"],
                    default=inspect.Parameter.empty,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                )
            )

    def test_generator_function(self) -> None:
        def gen() -> typing.Generator[str, None, None]:
            yield "abc"

        with pytest.raises(DependencyError, match="Generators are not supported"):
            create_dependency_from_parameter(
                inspect.Parameter(
                    name="dep",
                    annotation=typing.Annotated[str, gen],
                    default=inspect.Parameter.empty,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                )
            )

    def test_async_generator_function(self) -> None:
        async def gen() -> typing.AsyncGenerator[str, None]:
            yield "abc"

        with pytest.raises(DependencyError, match="Generators are not supported"):
            create_dependency_from_parameter(
                inspect.Parameter(
                    name="dep",
                    annotation=typing.Annotated[str, gen],
                    default=inspect.Parameter.empty,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                )
            )

    def test_type_alias(self) -> None:
        type Alias = typing.Annotated[str, VariableResolver("abc")]
        injection = create_dependency_from_parameter(
            inspect.Parameter(
                name="dep",
                default=inspect.Parameter.empty,
                annotation=Alias,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )
        assert injection.argument.param_type is str
        assert injection.argument.nullable is False
        assert isinstance(injection.resolver, VariableResolver)

    async def test_type_alias_generic(self) -> None:
        class MyModel: ...

        async def _from_json_resolver(arg: Argument) -> object:
            return arg.param_type

        type FromJSON[S] = typing.Annotated[S, FactoryResolver(_from_json_resolver)]

        injection = create_dependency_from_parameter(
            inspect.Parameter(
                name="dep",
                annotation=FromJSON[MyModel],
                default=inspect.Parameter.empty,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )
        context = ResolveContext(
            dependencies=Dependencies(
                {
                    HTTPConnection: VariableResolver(Request({"type": "http"})),
                    Argument: VariableResolver(injection.argument),
                }
            ),
        )
        assert injection.argument.param_type is MyModel
        assert await context.resolve(injection.resolver) == MyModel


class TestFromPath:
    async def test_from_path_no_type(self) -> None:
        context = ResolveContext(
            dependencies=Dependencies(
                {
                    HTTPConnection: VariableResolver(Request({"type": "http", "path_params": {"path": "abc"}})),
                    Argument: VariableResolver(
                        Argument(param_name="path", param_type=str, default=None, annotation=str, nullable=False)
                    ),
                }
            )
        )
        async with context:
            assert await context.resolve(FromPath) == "abc"

    async def test_from_path_with_type(self) -> None:
        context = ResolveContext(
            dependencies=Dependencies(
                {
                    HTTPConnection: VariableResolver(Request({"type": "http", "path_params": {"path": "42"}})),
                    Argument: VariableResolver(
                        Argument(param_name="path", param_type=int, default=None, annotation=str, nullable=False)
                    ),
                }
            )
        )
        async with context:
            assert await context.resolve(FromPath[int]) == 42



from __future__ import annotations

import functools
import typing

from anyio.to_thread import run_sync
from passlib.context import CryptContext


class PasswordHasher:
    def __init__(
        self,
        schemes: typing.Sequence[str] | None = None,
        default: str | None = None,
        deprecated: typing.Sequence[str] | typing.Literal["auto"] | None = None,
    ) -> None:
        schemes = schemes or ["pbkdf2_sha256"]
        self.context = CryptContext(
            schemes=schemes,
            default=default or schemes[0],
            deprecated=deprecated or "auto",
        )

    def make(self, plain_password: str, *, scheme: str | None = None) -> str:
        """Hash a plain password."""
        return self.context.hash(plain_password, scheme=scheme)

    def verify(
        self,
        plain_password: str,
        hashed_password: str,
        *,
        scheme: str | None = None,
    ) -> bool:
        """Verify a plain password against a hashed password."""
        return self.context.verify(plain_password, hashed_password, scheme=scheme)

    def verify_and_migrate(
        self,
        plain_password: str,
        hashed_password: str,
        *,
        scheme: str | None = None,
    ) -> tuple[bool, str | None]:
        """Verify password and re-hash the password if needed, all in a single call.
        Returns a tuple of (valid, new_hash)."""
        return self.context.verify_and_update(plain_password, hashed_password, scheme=scheme)

    def needs_update(self, hashed_password: str, *, scheme: str | None = None) -> bool:
        """Check if hash needs to be replaced for some reason,
        in which case the secret should be re-hashed."""
        return self.context.needs_update(hashed_password, scheme=scheme)

    async def amake(self, plain_password: str, *, scheme: str | None = None) -> str:
        """Hash a plain password."""
        return await run_sync(functools.partial(self.make, plain_password, scheme=scheme))

    async def averify(
        self,
        plain_password: str,
        hashed_password: str,
        *,
        scheme: str | None = None,
    ) -> bool:
        """Verify a plain password against a hashed password."""
        return await run_sync(functools.partial(self.verify, plain_password, hashed_password, scheme=scheme))

    async def averify_and_migrate(
        self,
        plain_password: str,
        hashed_password: str,
        *,
        scheme: str | None = None,
    ) -> tuple[bool, str | None]:
        """Verify password and re-hash the password if needed, all in a single call."""
        return await run_sync(
            functools.partial(
                self.verify_and_migrate,
                plain_password,
                hashed_password,
                scheme=scheme,
            )
        )



import json
import os
import typing as t

import jinja2
import pytest
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Route, Router
from starlette.testclient import TestClient

from kupala import Kupala, responses
from kupala.templating import Templates


class TestText:
    def test_response(self) -> None:
        http_response = responses.text("ok")
        client = TestClient(http_response)
        response = client.get("/")
        assert response.text == "ok"

    def test_with_status(self) -> None:
        http_response = responses.text("ok", status_code=201)
        client = TestClient(http_response)
        response = client.get("/")
        assert response.status_code == 201

    def test_with_headers(self) -> None:
        http_response = responses.text("ok", headers={"X-Header": "value"})
        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["x-header"] == "value"


class TestHTML:
    def test_response(self) -> None:
        http_response = responses.html("ok", status_code=201, headers={"X-Header": "value"})
        client = TestClient(http_response)
        response = client.get("/")
        assert response.status_code == 201
        assert response.headers["x-header"] == "value"
        assert response.headers["content-type"] == "text/html; charset=utf-8"
        assert response.text == "ok"

    def test_with_status(self) -> None:
        http_response = responses.html("ok", status_code=201)
        client = TestClient(http_response)
        response = client.get("/")
        assert response.status_code == 201

    def test_with_headers(self) -> None:
        http_response = responses.html("ok", headers={"X-Header": "value"})
        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["x-header"] == "value"


class TestEmpty:
    def test_response(self) -> None:
        http_response = responses.empty()
        client = TestClient(http_response)
        response = client.get("/")
        assert response.status_code == 204

    def test_with_headers(self) -> None:
        http_response = responses.empty(headers={"X-Header": "value"})
        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["x-header"] == "value"


class TestRedirect:
    def test_response(self) -> None:
        http_response = responses.redirect("/about")
        client = TestClient(http_response, follow_redirects=False)
        response = client.get("/")
        assert response.status_code == 302
        assert response.headers["location"] == "/about"

    def test_with_headers(self) -> None:
        http_response = responses.redirect("/about", headers={"X-Header": "value"})
        client = TestClient(http_response, follow_redirects=False)
        response = client.get("/")
        assert response.headers["x-header"] == "value"

    def test_with_code(self) -> None:
        http_response = responses.redirect("/about", status_code=307)
        client = TestClient(http_response, follow_redirects=False)
        response = client.get("/")
        assert response.status_code == 307


class TestRedirectToRoute:
    @pytest.fixture
    def http_request(self) -> Request:
        return Request(
            {
                "type": "http",
                "method": "GET",
                "url": "http://testserver/",
                "headers": [],
                "server": ("testserver", 80),
                "router": Router(
                    routes=[
                        Route("/about", endpoint=lambda r: Response(), name="about"),
                        Route("/page/{name}", endpoint=lambda r: Response(), name="page"),
                    ]
                ),
            }
        )

    def test_response(self, http_request: Request) -> None:
        http_response = responses.redirect_to_route(http_request, "about")
        client = TestClient(http_response, follow_redirects=False)
        response = client.get("/")
        assert response.status_code == 302
        assert response.headers["location"] == "http://testserver/about"

    def test_with_params(self, http_request: Request) -> None:
        http_response = responses.redirect_to_route(http_request, "page", {"name": "about"})
        client = TestClient(http_response, follow_redirects=False)
        response = client.get("/")
        assert response.headers["location"] == "http://testserver/page/about"

    def test_with_headers(self, http_request: Request) -> None:
        http_response = responses.redirect_to_route(http_request, "about", headers={"X-Header": "value"})
        client = TestClient(http_response, follow_redirects=False)
        response = client.get("/")
        assert response.headers["x-header"] == "value"
        assert response.headers["location"] == "http://testserver/about"

    def test_with_code(self, http_request: Request) -> None:
        http_response = responses.redirect_to_route(http_request, "about", status_code=307)
        client = TestClient(http_response, follow_redirects=False)
        response = client.get("/")
        assert response.status_code == 307
        assert response.headers["location"] == "http://testserver/about"


class TestJSON:
    def test_response(self) -> None:
        http_response = responses.json({"a": "b"})
        client = TestClient(http_response)
        response = client.get("/")
        assert response.text == '{\n  "a":"b"\n}'

    def test_with_status(self) -> None:
        http_response = responses.json({"a": "b"}, status_code=201)
        client = TestClient(http_response)
        response = client.get("/")
        assert response.status_code == 201

    def test_with_headers(self) -> None:
        http_response = responses.json({"a": "b"}, headers={"X-Header": "value"})
        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["x-header"] == "value"

    def test_with_indent(self) -> None:
        http_response = responses.json({"a": "b"}, indent=4)
        client = TestClient(http_response)
        response = client.get("/")
        assert response.text == '{\n    "a":"b"\n}'

    def test_with_encoder(self) -> None:
        class _Encoder(json.JSONEncoder):
            def default(self, o: object) -> object:
                return "[]"

        http_response = responses.json(set(), encoder_class=_Encoder)
        client = TestClient(http_response)
        response = client.get("/")
        assert response.text == '"[]"'


class TestJSONError:
    def test_response(self) -> None:
        http_response = responses.json_error("error")
        client = TestClient(http_response)
        response = client.get("/")
        assert response.json() == {"message": "error", "errors": {}}

    def test_with_errors(self) -> None:
        http_response = responses.json_error("error", errors={"a": ["b"]})
        client = TestClient(http_response)
        response = client.get("/")
        assert response.json() == {"message": "error", "errors": {"a": ["b"]}}

    def test_with_status(self) -> None:
        http_response = responses.json_error("error", status_code=201)
        client = TestClient(http_response)
        response = client.get("/")
        assert response.status_code == 201

    def test_with_headers(self) -> None:
        http_response = responses.json_error("error", headers={"X-Header": "value"})
        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["x-header"] == "value"


class TestSendFile:
    def test_response(self, tmpdir: os.PathLike[str]) -> None:
        file_path = os.path.join(tmpdir, "file.bin")
        with open(str(file_path), "wb") as f:
            f.write(b"content")

        http_response = responses.send_file(file_path)

        client = TestClient(http_response)
        response = client.get("/")
        assert response.content == b"content"
        assert response.headers["content-type"] == "application/octet-stream"

    def test_with_filename(self, tmpdir: os.PathLike[str]) -> None:
        file_path = os.path.join(tmpdir, "file.bin")
        with open(str(file_path), "wb") as f:
            f.write(b"content")

        http_response = responses.send_file(file_path, filename="file.bin")

        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["content-disposition"] == 'attachment; filename="file.bin"'

    def test_with_headers(self, tmpdir: os.PathLike[str]) -> None:
        file_path = os.path.join(tmpdir, "file.bin")
        with open(str(file_path), "wb") as f:
            f.write(b"content")

        http_response = responses.send_file(file_path, headers={"X-Header": "value"})

        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["x-header"] == "value"

    def test_with_media_type(self, tmpdir: os.PathLike[str]) -> None:
        file_path = os.path.join(tmpdir, "file.bin")
        with open(str(file_path), "wb") as f:
            f.write(b"content")

        http_response = responses.send_file(file_path, media_type="image/jpeg")

        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["content-type"] == "image/jpeg"

    def test_send_inline(self, tmpdir: os.PathLike[str]) -> None:
        file_path = os.path.join(tmpdir, "file.bin")
        with open(str(file_path), "wb") as f:
            f.write(b"content")

        http_response = responses.send_file(file_path, inline=True, filename="file.bin")

        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["content-disposition"] == 'inline; filename="file.bin"'


class TestStreaming:
    def test_response(self) -> None:
        async def content() -> t.AsyncGenerator[str, None]:
            yield "1"
            yield "2"
            yield "3"

        http_response = responses.stream(content())
        client = TestClient(http_response)
        response = client.get("/")
        assert response.text == "123"

    def test_with_headers(self) -> None:
        async def content() -> t.AsyncGenerator[str, None]:
            yield "1"
            yield "2"
            yield "3"

        http_response = responses.stream(content(), headers={"X-Header": "value"})
        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["x-header"] == "value"

    def test_with_status(self) -> None:
        async def content() -> t.AsyncGenerator[str, None]:
            yield "1"
            yield "2"
            yield "3"

        http_response = responses.stream(content(), status_code=201)
        client = TestClient(http_response)
        response = client.get("/")
        assert response.status_code == 201

    def test_with_media_type(self) -> None:
        async def content() -> t.AsyncGenerator[str, None]:
            yield "1"
            yield "2"
            yield "3"

        http_response = responses.stream(content(), media_type="image/jpeg")
        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["content-type"] == "image/jpeg"


class TestGoBack:
    @pytest.fixture
    def http_request(self) -> Request:
        return Request(
            {
                "type": "http",
                "method": "GET",
                "url": "http://testserver/",
                "headers": [],
                "server": ("testserver", 80),
                "path": "/",
                "router": Router(
                    routes=[
                        Route("/about", endpoint=lambda r: Response(), name="about"),
                    ]
                ),
            }
        )

    def test_response(self, http_request: Request) -> None:
        http_request.scope["headers"].append((b"referer", b"http://testserver/somepage"))
        http_response = responses.back(http_request)
        client = TestClient(http_response, follow_redirects=False)
        response = client.get("/")
        assert response.status_code == 302
        assert response.headers["location"] == "http://testserver/somepage"

    def test_bad_referer(self, http_request: Request) -> None:
        http_request.scope["headers"].append((b"referer", b"http://example.com/somepage"))
        http_response = responses.back(http_request)
        client = TestClient(http_response, follow_redirects=False)
        response = client.get("/")
        assert response.headers["location"] == "http://testserver/"

    def test_no_referer(self, http_request: Request) -> None:
        http_response = responses.back(http_request)
        client = TestClient(http_response, follow_redirects=False)
        response = client.get("/")
        assert response.headers["location"] == "http://testserver/"

    def test_with_code(self, http_request: Request) -> None:
        http_response = responses.back(http_request, status_code=307)
        client = TestClient(http_response, follow_redirects=False)
        response = client.get("/")
        assert response.status_code == 307

    def test_with_headers(self, http_request: Request) -> None:
        http_response = responses.back(http_request, headers={"X-Header": "value"})
        client = TestClient(http_response, follow_redirects=False)
        response = client.get("/")
        assert response.headers["x-header"] == "value"


class TestTemplate:
    @pytest.fixture
    def http_request(self) -> Request:
        return Request(
            {
                "type": "http",
                "method": "GET",
                "url": "http://testserver/",
                "headers": [],
                "server": ("testserver", 80),
                "path": "/",
                "router": Router(
                    routes=[
                        Route("/about", endpoint=lambda r: Response(), name="about"),
                    ]
                ),
                "app": Kupala(
                    secret_key="key!",
                    extensions=[
                        Templates(
                            template_loaders=[
                                jinja2.DictLoader(
                                    {
                                        "index.html": "Hello, {{ name }}!",
                                        "nocontext.html": "Hello!",
                                    }
                                )
                            ]
                        ),
                    ],
                ),
            }
        )

    def test_response(self, http_request: Request) -> None:
        http_response = responses.template(http_request, "nocontext.html")
        client = TestClient(http_response)
        response = client.get("/")
        assert response.text == "Hello!"

    def test_with_context(self, http_request: Request) -> None:
        http_response = responses.template(http_request, "index.html", {"name": "world"})
        client = TestClient(http_response)
        response = client.get("/")
        assert response.text == "Hello, world!"

    def test_with_code(self, http_request: Request) -> None:
        http_response = responses.template(http_request, "index.html", {"name": "world"}, status_code=201)
        client = TestClient(http_response)
        response = client.get("/")
        assert response.status_code == 201

    def test_with_headers(self, http_request: Request) -> None:
        http_response = responses.template(http_request, "index.html", {"name": "world"}, headers={"X-Header": "value"})
        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["x-header"] == "value"

    def test_with_media_type(self, http_request: Request) -> None:
        http_response = responses.template(http_request, "index.html", {"name": "world"}, media_type="image/jpeg")
        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["content-type"] == "image/jpeg"


class TestTemplateBlock:
    @pytest.fixture
    def http_request(self) -> Request:
        return Request(
            {
                "type": "http",
                "method": "GET",
                "url": "http://testserver/",
                "headers": [],
                "server": ("testserver", 80),
                "path": "/",
                "router": Router(
                    routes=[
                        Route("/about", endpoint=lambda r: Response(), name="about"),
                    ]
                ),
                "app": Kupala(
                    secret_key="key!",
                    extensions=[
                        Templates(
                            template_loaders=[
                                jinja2.DictLoader(
                                    {
                                        "index.html": "{% block content %}Hello, {{ name }}!{% endblock %}",
                                        "nocontext.html": "{% block content %}Hello!{% endblock %}",
                                    }
                                )
                            ]
                        ),
                    ],
                ),
            }
        )

    def test_response(self, http_request: Request) -> None:
        http_response = responses.template_block(http_request, "nocontext.html", "content")
        client = TestClient(http_response)
        response = client.get("/")
        assert response.text == "Hello!"

    def test_with_context(self, http_request: Request) -> None:
        http_response = responses.template_block(http_request, "index.html", "content", {"name": "world"})
        client = TestClient(http_response)
        response = client.get("/")
        assert response.text == "Hello, world!"

    def test_with_code(self, http_request: Request) -> None:
        http_response = responses.template_block(
            http_request, "index.html", "content", {"name": "world"}, status_code=201
        )
        client = TestClient(http_response)
        response = client.get("/")
        assert response.status_code == 201

    def test_with_headers(self, http_request: Request) -> None:
        http_response = responses.template_block(
            http_request, "index.html", "content", {"name": "world"}, headers={"X-Header": "value"}
        )
        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["x-header"] == "value"

    def test_with_media_type(self, http_request: Request) -> None:
        http_response = responses.template_block(
            http_request, "index.html", "content", {"name": "world"}, media_type="image/jpeg"
        )
        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["content-type"] == "image/jpeg"


class TestTemplateMacro:
    @pytest.fixture
    def http_request(self) -> Request:
        return Request(
            {
                "type": "http",
                "method": "GET",
                "url": "http://testserver/",
                "headers": [],
                "server": ("testserver", 80),
                "path": "/",
                "router": Router(
                    routes=[
                        Route("/about", endpoint=lambda r: Response(), name="about"),
                    ]
                ),
                "app": Kupala(
                    secret_key="key!",
                    extensions=[
                        Templates(
                            template_loaders=[
                                jinja2.DictLoader(
                                    {
                                        "index.html": "{% macro content(name) %}Hello, {{ name }}!{% endmacro %}",
                                        "nocontext.html": "{% macro content(name) %}Hello!{% endmacro %}",
                                    }
                                )
                            ]
                        ),
                    ],
                ),
            }
        )

    def test_response(self, http_request: Request) -> None:
        http_response = responses.template_macro(http_request, "nocontext.html", "content")
        client = TestClient(http_response)
        response = client.get("/")
        assert response.text == "Hello!"

    def test_with_context(self, http_request: Request) -> None:
        http_response = responses.template_macro(http_request, "index.html", "content", {"name": "world"})
        client = TestClient(http_response)
        response = client.get("/")
        assert response.text == "Hello, world!"

    def test_with_code(self, http_request: Request) -> None:
        http_response = responses.template_macro(
            http_request, "index.html", "content", {"name": "world"}, status_code=201
        )
        client = TestClient(http_response)
        response = client.get("/")
        assert response.status_code == 201

    def test_with_headers(self, http_request: Request) -> None:
        http_response = responses.template_macro(
            http_request, "index.html", "content", {"name": "world"}, headers={"X-Header": "value"}
        )
        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["x-header"] == "value"

    def test_with_media_type(self, http_request: Request) -> None:
        http_response = responses.template_macro(
            http_request, "index.html", "content", {"name": "world"}, media_type="image/jpeg"
        )
        client = TestClient(http_response)
        response = client.get("/")
        assert response.headers["content-type"] == "image/jpeg"



import pytest

from kupala.exceptions import (
    ErrorCode,
    BadRequest,
    Conflict,
    HTTPException,
    KupalaError,
    MethodNotAllowed,
    NotAcceptable,
    NotAuthenticated,
    NotAuthorized,
    NotFound,
    RequestTimeout,
    TooManyRequests,
    UnprocessableEntity,
    UnsupportedMediaType,
    ValidationError,
)


class TestErrorCode:
    def test_str(self) -> None:
        code = ErrorCode("test", "Test error")
        assert str(code) == "Test error"

    def test_eq(self) -> None:
        code = ErrorCode("test", "Test error")
        assert code == "test"
        assert code == ErrorCode("test", "Test error")
        assert code != "unknown"
        assert code != ErrorCode("unknown", "Unknown error")

        with pytest.raises(NotImplementedError):
            assert code == object()

    def test_int_eq(self) -> None:
        code = ErrorCode[int](100, "Test error")
        assert code == 100
        assert code == ErrorCode[int](100, "Test error")
        assert code != 200
        assert code != ErrorCode[int](200, "Unknown error")


class TestKupalaError:
    def test_simple(self) -> None:
        with pytest.raises(KupalaError, match=""):
            raise KupalaError

    def test_with_message(self) -> None:
        with pytest.raises(KupalaError, match="error"):
            raise KupalaError("error")

    def test_with_error_code_message(self) -> None:
        code = ErrorCode("unknown_error", "Unknown error")
        with pytest.raises(KupalaError, match="Unknown error"):
            raise KupalaError(error_code=code)

    def test_with_message_overrides_error_code(self) -> None:
        code = ErrorCode("unknown_error", "Unknown error")
        with pytest.raises(KupalaError, match="custom error"):
            raise KupalaError("custom error", error_code=code)

    def test_with_message_from_attr(self) -> None:
        class Error(KupalaError):
            message = "Custom error"

        with pytest.raises(KupalaError, match="Custom error"):
            raise Error

    def test_message_from_attr_error_code(self) -> None:
        class Error(KupalaError):
            error_code = ErrorCode("unknown_error", "Unknown error")

        with pytest.raises(KupalaError, match="Unknown error") as ex:
            raise Error
        assert ex.value.error_code == ErrorCode("unknown_error", "Unknown error")

    def test_with_attr_code_override_attr_message(self) -> None:
        class Error(KupalaError):
            message = "Custom error"
            error_code = ErrorCode("unknown_error", "Code error")

        with pytest.raises(KupalaError, match="Code error"):
            raise Error


class TestHTTPError:
    def test_simple(self) -> None:
        exc = HTTPException(status_code=400)
        assert exc.status_code == 400
        assert exc.detail == "Bad Request"

    def test_detail(self) -> None:
        exc = HTTPException("Custom error", status_code=400)
        assert exc.status_code == 400
        assert exc.detail == "Custom error"

    def test_headers(self) -> None:
        exc = HTTPException("Custom error", status_code=400, headers={"X-Test": "test"})
        assert exc.status_code == 400
        assert exc.detail == "Custom error"
        assert exc.headers == {"X-Test": "test"}

    def test_error_code(self) -> None:
        code = ErrorCode("unknown_error", "Unknown error")
        exc = HTTPException("Custom error", status_code=400, headers={"X-Test": "test"}, error_code=code)
        assert exc.status_code == 400
        assert exc.error_code == code
        assert exc.detail == "Custom error"
        assert exc.headers == {"X-Test": "test"}

    def test_validation_error(self) -> None:
        exc = ValidationError(status_code=400, errors={"field": ["error"]})
        assert exc.errors == {"field": ["error"]}


@pytest.mark.parametrize(
    "status_code, exc_class, message",
    [
        (400, BadRequest, "Bad Request"),
        (401, NotAuthenticated, "Unauthorized"),
        (403, NotAuthorized, "Forbidden"),
        (404, NotFound, "Not Found"),
        (405, MethodNotAllowed, "Method Not Allowed"),
        (406, NotAcceptable, "Not Acceptable"),
        (408, RequestTimeout, "Request Timeout"),
        (409, Conflict, "Conflict"),
        (415, UnsupportedMediaType, "Unsupported Media Type"),
        (422, UnprocessableEntity, "Unprocessable Content"),
        (422, ValidationError, "Unprocessable Content"),
        (429, TooManyRequests, "Too Many Requests"),
    ],
)
def test_standard_exceptions(status_code: int, exc_class: type[HTTPException], message: str) -> None:
    exc = exc_class()
    assert exc.status_code == status_code
    assert exc.detail == message



import typing

import pytest
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, async_sessionmaker, create_async_engine
from starlette.applications import Starlette
from starlette.routing import BaseRoute
from starlette.testclient import TestClient
from starlette.types import ASGIApp

from kupala.database.models import Base
from kupala.routing import Middleware, RouteGroup
from tests.models import Profile, User


class AppFactory(typing.Protocol):  # pragma: nocover
    def __call__(
        self,
        debug: bool = True,
        middleware: list[Middleware] | None = None,
        routes: typing.Iterable[BaseRoute] | None = None,
        **kwargs: typing.Any,
    ) -> Starlette: ...


class ClientFactory(typing.Protocol):  # pragma: nocover
    def __call__(
        self,
        debug: bool = True,
        middleware: list[Middleware] | None = None,
        routes: typing.Iterable[BaseRoute] | None = None,
        raise_server_exceptions: bool = True,
        app: ASGIApp | None = None,
        **kwargs: typing.Any,
    ) -> TestClient: ...


@pytest.fixture
def test_app_factory() -> AppFactory:
    def factory(*args: typing.Any, **kwargs: typing.Any) -> Starlette:
        kwargs.setdefault("debug", True)
        kwargs.setdefault("routes", RouteGroup())
        kwargs.setdefault("middleware", [])
        return Starlette(*args, **kwargs)

    return factory


@pytest.fixture
def test_client_factory(test_app_factory: AppFactory) -> ClientFactory:
    def factory(**kwargs: typing.Any) -> TestClient:
        raise_server_exceptions = kwargs.pop("raise_server_exceptions", True)
        app = kwargs.pop("app", test_app_factory(**kwargs))
        return TestClient(app, raise_server_exceptions=raise_server_exceptions)

    return typing.cast(ClientFactory, factory)


@pytest.fixture
def routes() -> RouteGroup:
    return RouteGroup()


@pytest.fixture()
async def dbengine() -> typing.AsyncGenerator[AsyncEngine, None]:
    engine = create_async_engine("sqlite+aiosqlite:///:memory:")
    async with engine.connect() as conn:
        await conn.run_sync(Base.metadata.drop_all)
        await conn.run_sync(Base.metadata.create_all)
        yield engine


@pytest.fixture
async def dbsession(dbengine: AsyncEngine) -> typing.AsyncGenerator[AsyncSession, None]:
    async with dbengine.connect() as conn:
        async with conn.begin() as tx:
            async with async_sessionmaker(
                bind=conn,
                expire_on_commit=False,
                join_transaction_mode="create_savepoint",
            )() as dbsession:
                try:
                    yield dbsession
                finally:
                    await tx.rollback()


@pytest.fixture(autouse=True)
async def setup_users(dbsession: AsyncSession) -> None:
    users = [
        User(id=1, name="user_01", email="01@user", password="", profile=Profile(bio="bio_01")),
        User(id=2, name="user_02", email="02@user", password="", profile=Profile(bio="bio_02")),
        User(id=3, name="user_03", email="03@user", password="", profile=Profile(bio="bio_03")),
        User(id=4, name="user_04", email="04@user", password="", profile=Profile(bio="bio_04")),
        User(id=5, name="user_05", email="05@user", password="", profile=Profile(bio="bio_05")),
        User(id=6, name="user_06", email="06@user", password="", profile=Profile(bio="bio_06")),
        User(id=7, name="user_07", email="07@user", password="", profile=Profile(bio="bio_07")),
        User(id=8, name="user_08", email="08@user", password="", profile=Profile(bio="bio_08")),
        User(id=9, name="user_09", email="09@user", password="", profile=Profile(bio="bio_09")),
    ]
    dbsession.add_all(users)
    await dbsession.flush()


@pytest.fixture()
def user() -> User:
    return User(email="root@localhost")



from __future__ import annotations

import pytest
import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession

from kupala.database.query import MultipleResultsError, NoResultError, query
from tests.models import User


class TestOne:
    async def test_one(self, dbsession: AsyncSession) -> None:
        stmt = sa.select(User).where(User.id == 1)
        model = await query(dbsession).select(stmt).one()
        assert model.id == 1

    async def test_one_no_rows(self, dbsession: AsyncSession) -> None:
        with pytest.raises(NoResultError):
            stmt = sa.select(User).where(User.id == -1)
            await query(dbsession).select(stmt).one()

    async def test_one_multiple_rows(self, dbsession: AsyncSession) -> None:
        with pytest.raises(MultipleResultsError):
            stmt = sa.select(User).where(sa.or_(User.id == 1, User.id == 2))
            await query(dbsession).select(stmt).one()


class TestOneOrNone:
    async def test_one_or_none(self, dbsession: AsyncSession) -> None:
        stmt = sa.select(User).where(User.id == 1)
        model = await query(dbsession).select(stmt).one_or_none()
        assert model is not None
        assert model.id == 1

    async def test_one_or_none_no_row(self, dbsession: AsyncSession) -> None:
        stmt = sa.select(User).where(User.id == -1)
        model = await query(dbsession).select(stmt).one_or_none()
        assert model is None

    async def test_one_or_none_multiple_rows(self, dbsession: AsyncSession) -> None:
        with pytest.raises(MultipleResultsError):
            stmt = sa.select(User).where(sa.or_(User.id == 1, User.id == 2))
            await query(dbsession).select(stmt).one_or_none()


class TestOneOrRaise:
    async def test_one_or_raise(self, dbsession: AsyncSession) -> None:
        stmt = sa.select(User).where(User.id == 1)
        model = await query(dbsession).select(stmt).one_or_raise(ValueError())
        assert model

    async def test_one_or_raise_no_row(self, dbsession: AsyncSession) -> None:
        with pytest.raises(ValueError):
            await query(dbsession).select(sa.select(User).where(User.id == -1)).one_or_raise(ValueError())

    async def test_one_or_raise_multiple_rows(self, dbsession: AsyncSession) -> None:
        with pytest.raises(MultipleResultsError):
            stmt = sa.select(User).where(sa.or_(User.id == 1, User.id == 2))
            await query(dbsession).select(stmt).one_or_raise(ValueError())


class TestOneOrDefault:
    async def test_one_or_default(self, dbsession: AsyncSession) -> None:
        stmt = sa.select(User).where(User.id == 1)
        model = await query(dbsession).select(stmt).one_or_default(User(id=-1, name="n/a", email="n/a"))
        assert model.id == 1

    async def test_one_or_default_no_row(self, dbsession: AsyncSession) -> None:
        stmt = sa.select(User).where(User.id == -1)
        model = await query(dbsession).select(stmt).one_or_default(User(id=-1, name="n/a", email="n/a"))
        assert model.id == -1


async def test_of(dbsession: AsyncSession) -> None:
    stmt = query.of(User)
    models = await query(dbsession).select(stmt).all()
    assert len(models) == 9


async def test_all(dbsession: AsyncSession) -> None:
    stmt = sa.select(User)
    models = await query(dbsession).select(stmt).all()
    assert len(models) == 9


async def test_get(dbsession: AsyncSession) -> None:
    assert await query(dbsession).get(User, 1)
    assert not await query(dbsession).get(User, -1)


async def test_iterator(dbsession: AsyncSession) -> None:
    stmt = sa.select(User).limit(3)
    iterator = query(dbsession).select(stmt).iterator(batch_size=1)
    assert [1, 2, 3] == [model.id async for model in iterator]


async def test_exists(dbsession: AsyncSession) -> None:
    stmt = sa.select(User).where(User.id == 1)
    assert await query(dbsession).select(stmt).exists() is True

    stmt = sa.select(User).where(User.id == -1)
    assert await query(dbsession).select(stmt).exists() is False


async def test_count(dbsession: AsyncSession) -> None:
    stmt = sa.select(User)
    assert await query(dbsession).select(stmt).count() == 9

    stmt = sa.select(User).where(User.id == 1)
    assert await query(dbsession).select(stmt).count() == 1


class TestChoices:
    async def test_choices(self, dbsession: AsyncSession) -> None:
        stmt = sa.select(User).limit(3)
        choices = [choice async for choice in query(dbsession).select(stmt).choices()]
        assert choices == [(1, "01@user"), (2, "02@user"), (3, "03@user")]

    async def test_choices_string_keys(self, dbsession: AsyncSession) -> None:
        stmt = sa.select(User).limit(3)
        choices: list[tuple[int, str]] = [
            choice async for choice in query(dbsession).select(stmt).choices(label_attr="name", value_attr="id")
        ]
        assert choices == [(1, "user_01"), (2, "user_02"), (3, "user_03")]

    async def test_choices_callable_keys(self, dbsession: AsyncSession) -> None:
        stmt = sa.select(User).limit(3)
        choices = [
            choice
            async for choice in query(dbsession)
            .select(stmt)
            .choices(label_attr=lambda obj: obj.name, value_attr=lambda o: o.id)
        ]
        assert choices == [(1, "user_01"), (2, "user_02"), (3, "user_03")]



from unittest import mock

from starlette.types import Receive, Scope, Send

from kupala.database.manager import Database
from kupala.database.middleware import DbSessionMiddleware


empty_send = mock.AsyncMock()
empty_receive = mock.AsyncMock()


async def test_injects_dbsession() -> None:
    async def app(scope: Scope, receive: Receive, send: Send) -> None:
        pass

    middleware = DbSessionMiddleware(
        app,
        {
            "default": Database("sqlite+aiosqlite:///:memory:"),
        },
    )
    scope: Scope = {"type": "http"}
    await middleware(scope, empty_receive, empty_send)
    assert "dbsession" in scope["state"]
    assert "dbsession_default" in scope["state"]


async def test_injects_multiple_dbsessions() -> None:
    async def app(scope: Scope, receive: Receive, send: Send) -> None:
        pass

    middleware = DbSessionMiddleware(
        app,
        {
            "default": Database("sqlite+aiosqlite:///:memory:"),
            "read_only": Database("sqlite+aiosqlite:///:memory:"),
        },
    )
    scope: Scope = {"type": "http"}
    await middleware(scope, empty_receive, empty_send)
    assert "dbsession" in scope["state"]
    assert "dbsession_default" in scope["state"]
    assert "dbsession_read_only" in scope["state"]



import io
import os.path

import jinja2
import pytest
from async_storages import FileStorage, MemoryBackend
from starlette.datastructures import UploadFile
from starlette.requests import Request
from starlette.responses import Response
from starlette.testclient import TestClient

from kupala import Kupala
from kupala.files import Files, StaticFiles, static_url
from kupala.routing import Route
from kupala.templating import Templates


class TestStaticFiles:
    def test_static_files(self) -> None:
        app = Kupala(
            secret_key="key!",
            extensions=[
                StaticFiles(packages=[("tests", "assets")]),
            ],
        )
        with TestClient(app) as client:
            response = client.get("/static/somefile.txt")
            assert response.status_code == 200
            assert response.headers["content-type"] == "text/plain; charset=utf-8"
            assert response.text == "somecontent\n"

    def test_static_files_directory(self) -> None:
        this_dir = os.path.dirname(__file__)
        app = Kupala(
            secret_key="key!",
            extensions=[
                StaticFiles(directory=os.path.join(this_dir, "assets")),
            ],
        )
        with TestClient(app) as client:
            response = client.get("/static/somefile.txt")
            assert response.status_code == 200
            assert response.headers["content-type"] == "text/plain; charset=utf-8"
            assert response.text == "somecontent\n"

    def test_static_files_default_route_name(self) -> None:
        app = Kupala(
            secret_key="key!",
            extensions=[
                StaticFiles(packages=[("tests", "assets")]),
            ],
        )
        url = app.url_path_for("static", path="somefile.txt")
        assert url == "/static/somefile.txt"

    def test_static_files_route_name(self) -> None:
        app = Kupala(
            secret_key="key!",
            extensions=[
                StaticFiles(route_name="assets", packages=[("tests", "assets")]),
            ],
        )
        url = app.url_path_for("assets", path="somefile.txt")
        assert url == "/static/somefile.txt"

    def test_static_files_default_url_prefix(self) -> None:
        app = Kupala(
            secret_key="key!",
            extensions=[
                StaticFiles(packages=[("tests", "assets")]),
            ],
        )
        url = app.url_path_for("static", path="somefile.txt")
        assert url == "/static/somefile.txt"

    def test_static_files_url_prefix(self) -> None:
        app = Kupala(
            secret_key="key!",
            extensions=[
                StaticFiles(url_prefix="/assets", packages=[("tests", "assets")]),
            ],
        )
        url = app.url_path_for("static", path="somefile.txt")
        assert url == "/assets/somefile.txt"

    def test_template_helper(self) -> None:
        async def view(request: Request) -> Response:
            return Templates.of(request.app).render_to_response(request, "index.html")

        app = Kupala(
            secret_key="key!",
            routes=[Route("/", view)],
            extensions=[
                Templates(
                    template_loaders=[jinja2.DictLoader({"index.html": "{{ static_url('somefile.txt') }}"})],
                ),
                StaticFiles(url_prefix="/assets", packages=[("tests", "assets")]),
            ],
        )
        with TestClient(app) as client:
            response = client.get("/")
            assert response.status_code == 200
            assert response.text.startswith("http://testserver/assets/somefile.txt")


class TestStaticUrl:
    @pytest.fixture
    def http_request(self) -> Request:
        return Request(
            {
                "type": "http",
                "method": "GET",
                "url": "http://testserver/",
                "headers": [],
                "server": ("testserver", 80),
                "path": "/",
                "app": Kupala(secret_key="key!", extensions=[StaticFiles(packages=[("tests", "assets")])]),
            }
        )

    def test_generates_url(self, http_request: Request) -> None:
        url = static_url(http_request, "/image.jpg")
        assert url.path == "/static/image.jpg"

    def test_cache_prefix_no_debug(self, http_request: Request) -> None:
        http_request.app.debug = False
        url = static_url(http_request, "/image.jpg")
        url2 = static_url(http_request, "/image.jpg")
        assert url == url2

    def test_cache_prefix_debug(self, http_request: Request) -> None:
        http_request.app.debug = True
        url = static_url(http_request, "/image.jpg")
        url2 = static_url(http_request, "/image.jpg")
        assert url != url2

    @pytest.mark.parametrize("image_url", ["http://example.com/image.jpg", "https://example.com/image.jpg"])
    def test_ignores_http(self, http_request: Request, image_url: str) -> None:
        url = static_url(http_request, image_url)
        assert url == url


class TestFiles:
    async def test_serves_files(self) -> None:
        files = Files(
            default="memory",
            storages={
                "memory": FileStorage(MemoryBackend()),
            },
        )
        app = Kupala(
            secret_key="key!",
            extensions=[files],
        )
        await files.write("somefile.txt", b"somecontent\n")
        with TestClient(app) as client:
            response = client.get("/media/somefile.txt")
            assert response.status_code == 200
            assert response.headers["content-type"] == "text/plain; charset=utf-8"
            assert response.text == "somecontent\n"

    async def test_serves_files_custom_route_name(self) -> None:
        files = Files(
            default="memory",
            route_name="files",
            storages={
                "memory": FileStorage(MemoryBackend()),
            },
        )
        app = Kupala(
            secret_key="key!",
            extensions=[files],
        )
        url = app.url_path_for("files", path="somefile.txt")
        assert url == "/media/somefile.txt"

    async def test_serves_files_custom_url_prefix(self) -> None:
        files = Files(
            default="memory",
            url_prefix="/files",
            storages={
                "memory": FileStorage(MemoryBackend()),
            },
        )
        app = Kupala(
            secret_key="key!",
            extensions=[files],
        )
        url = app.url_path_for("media", path="somefile.txt")
        assert url == "/files/somefile.txt"

    async def test_serves_files_custom_default(self) -> None:
        files = Files(
            default="second",
            url_prefix="/files",
            storages={
                "memory": FileStorage(MemoryBackend()),
                "second": FileStorage(MemoryBackend()),
            },
        )
        app = Kupala(
            secret_key="key!",
            extensions=[files],
        )
        url = app.url_path_for("media", path="somefile.txt")
        assert url == "/files/somefile.txt"

    async def test_serves_files_invalid_default(self) -> None:
        with pytest.raises(AssertionError):
            Files(
                default="missing",
                url_prefix="/files",
                storages={
                    "memory": FileStorage(MemoryBackend()),
                },
            )


class TestFilesMethodsIntegration:
    @pytest.fixture
    async def files_instance(self) -> Files:
        return Files(
            default="memory",
            storages={"memory": FileStorage(MemoryBackend())},
        )

    async def test_write_and_exists(self, files_instance: Files) -> None:
        await files_instance.write("test.txt", b"test content")
        assert await files_instance.exists("test.txt")
        assert not await files_instance.exists("nonexistent.txt")

    async def test_open_and_read(self, files_instance: Files) -> None:
        await files_instance.write("test.txt", b"test content")
        async with await files_instance.open("test.txt") as file:
            content = await file.read()
            assert content == b"test content"

    async def test_readlines(self, files_instance: Files) -> None:
        await files_instance.write("test.txt", b"line1\nline2\nline3")
        lines = []
        async for line in await files_instance.readlines("test.txt"):
            lines.append(line)
        assert lines == [b"line1\n", b"line2\n", b"line3"]

    async def test_delete(self, files_instance: Files) -> None:
        await files_instance.write("test.txt", b"test content")
        assert await files_instance.exists("test.txt")
        await files_instance.delete("test.txt")
        assert not await files_instance.exists("test.txt")

    async def test_delete_many(self, files_instance: Files) -> None:
        await files_instance.write("file1.txt", b"content1")
        await files_instance.write("file2.txt", b"content2")
        assert await files_instance.exists("file1.txt")
        assert await files_instance.exists("file2.txt")
        await files_instance.delete_many(["file1.txt", "file2.txt"])
        assert not await files_instance.exists("file1.txt")
        assert not await files_instance.exists("file2.txt")

    async def test_upload(self, files_instance: Files) -> None:
        upload_file = UploadFile(io.BytesIO(b"test content"), filename="test.txt")

        file_path = await files_instance.upload(upload_file, "uploads")
        assert await files_instance.exists(file_path)
        async with await files_instance.open(file_path) as file:
            content = await file.read()
            assert content == b"test content"

    async def test_get_storage(self, files_instance: Files) -> None:
        storage = files_instance.get_storage("memory")
        assert isinstance(storage, FileStorage)

    async def test_of_method(self) -> None:
        files = Files(
            default="memory",
            storages={"memory": FileStorage(MemoryBackend())},
        )
        app = Kupala(
            secret_key="key!",
            extensions=[files],
        )
        assert Files.of(app) is files



import contextlib
import contextvars
import inspect
import typing
import asyncio

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)


class NoActiveSession(Exception): ...


EngineOptions = dict[str, typing.Any]
SessionOptions = dict[str, typing.Any]


class Database:
    def __init__(
        self,
        url: str,
        *,
        echo: bool = False,
        session_class: type[AsyncSession] = AsyncSession,
        engine_options: EngineOptions | None = None,
        session_options: SessionOptions | None = None,
    ) -> None:
        self._engine = create_async_engine(
            url=url,
            echo=echo,
            **(engine_options or {}),
        )

        _session_options: dict[str, typing.Any] = {
            "expire_on_commit": False,
            "class_": session_class,
        }
        _session_options.update(session_options or {})
        self._sessionmaker = async_sessionmaker(bind=self._engine, **_session_options)
        self._current_session: contextvars.ContextVar[AsyncSession] = contextvars.ContextVar("sqla_current_session")

    async def __aenter__(self) -> AsyncEngine:
        return self._engine

    async def __aexit__(self, *exc: typing.Any) -> None:
        pass

    @property
    def current_session(self) -> AsyncSession | None:
        """Return the currently active session if it exists.
        Prefer not to use this method, only for debugging and special purposes.
        Raises ValueError if no session is active."""
        try:
            return self._current_session.get()
        except LookupError:
            raise NoActiveSession("No SQLAlchemy session is currently active.")

    @contextlib.asynccontextmanager
    async def session(self, force_rollback: bool = False) -> typing.AsyncGenerator[AsyncSession, None]:
        """Get a new session instance.
        If force_rollback is True, the session will be rolled back after exiting the context."""

        factory: typing.Callable[[], typing.AsyncContextManager[AsyncSession]] = self._start_normal_session
        if force_rollback:
            factory = self._start_rollback_session

        async with factory() as session:
            restore_token = self._current_session.set(session)
            try:
                yield session
            finally:
                self._current_session.reset(restore_token)

    @contextlib.asynccontextmanager
    async def _start_normal_session(self, **kwargs: typing.Any) -> typing.AsyncGenerator[AsyncSession, None]:
        async with self._sessionmaker(**kwargs) as session:
            yield session

    @contextlib.asynccontextmanager
    async def _start_rollback_session(
        self,
    ) -> typing.AsyncGenerator[AsyncSession, None]:
        """Create a new session that will be rolled back after exiting the context.
        For testing purposes. Any BEGIN/COMMIT/ROLLBACK ops are executed in SAVEPOINT.
        See https://docs.sqlalchemy.org/en/20/orm/session_transaction.html#joining-a-session-into-an-external-transaction-such-as-for-test-suites"""
        async with self._engine.connect() as conn:
            async with conn.begin() as tx:
                async with self._start_normal_session(bind=conn, join_transaction_mode="create_savepoint") as session:
                    try:
                        yield session
                    finally:
                        await tx.rollback()


def on_commit(dbsession: AsyncSession, func: typing.Callable[..., typing.Any]) -> None:
    """Register a function to be called after the session is committed.
    The function will be called with the session as its only argument."""
    dbsession.info.setdefault("after_commit", []).append(func)

    @sa.event.listens_for(dbsession.sync_session, "after_commit")
    def _on_commit(session: AsyncSession) -> None:
        for func in session.info.pop("after_commit", []):
            if inspect.iscoroutinefunction(func):
                asyncio.get_event_loop().create_task(func())
            else:
                func()



import datetime
import enum
import typing
import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase, Mapped

from kupala.database.columns import AutoCreatedAt, AutoUpdatedAt


class WithTimestamps:
    __abstract__ = True
    created_at: Mapped[AutoCreatedAt]
    updated_at: Mapped[AutoUpdatedAt]


class Base(AsyncAttrs, DeclarativeBase):
    __abstract__ = True
    __repr_field__: str | None = None
    metadata = sa.MetaData()
    type_annotation_map: dict[typing.Any, typing.Any] = {
        datetime.datetime: sa.DateTime(timezone=True),
        enum.StrEnum: sa.Text(),
        enum.IntEnum: sa.Integer(),
    }

    def __str__(self) -> str:
        if hasattr(self, "name"):
            return str(self.name)

        if hasattr(self, "title"):
            return str(self.title)
        return "n/a"

    def __repr__(self) -> str:
        pk = [column.name for column in self.__class__.__mapper__.primary_key]
        attrs = self.__repr_field__ or pk
        values = [f'{attr}="{getattr(self, attr)}"' for attr in attrs]
        return f"{self.__class__.__name__}({', '.join(map(str, values))})"



import pathlib
import typing

import jinja2
import pytest
from starlette.requests import Request
from starlette.responses import PlainTextResponse, Response

from kupala import Kupala
from kupala.routing import Route
from kupala.templating import Templates

jinja_env = jinja2.Environment(
    loader=jinja2.DictLoader(
        {
            "index.html": "Hello, {{ name }}!",
            "macro.html": "{% macro hello(name) %}Hello, {{ name }}!{% endmacro %}",
            "block.html": "{% block content %}Hello, {{ name }}!{% endblock %}",
            "request.html": "{{ request }}",
            "app.html": "{{ app }}",
            "url_matches.html": "{{ url_matches('/test') }} - {{ url_matches('/no-test') }}",
            "route_matches.html": "{{ route_matches('test') }} - {{ route_matches('no-test') }}",
        }
    ),
)


def test_directories(tmp_path: pathlib.Path) -> None:
    (tmp_path / "index.html").write_text("Hello, world!")
    templates = Templates(directories=[tmp_path])
    assert templates.render("index.html") == "Hello, world!"


def test_template_loaders() -> None:
    templates = Templates(template_loaders=[jinja2.DictLoader({"index.html": "Hello, world!"})])
    assert templates.render("index.html") == "Hello, world!"


def test_globals() -> None:
    templates = Templates(
        template_loaders=[jinja2.DictLoader({"index.html": "Hello, {{ name }}!"})],
        globals={"name": "world"},
    )
    assert templates.render("index.html") == "Hello, world!"


def test_filters() -> None:
    templates = Templates(
        template_loaders=[jinja2.DictLoader({"index.html": "Hello, {{ name|up }}!"})],
        filters={"up": str.upper},
    )
    assert templates.render("index.html", {"name": "world"}) == "Hello, WORLD!"


def test_tests() -> None:
    templates = Templates(
        template_loaders=[jinja2.DictLoader({"index.html": "Hello, {{ 'UP' if name is up else 'up' }}!"})],
        tests={"up": lambda x: x.isupper()},
    )
    assert templates.render("index.html", {"name": "WORLD"}) == "Hello, UP!"


def test_undefined_on() -> None:
    templates = Templates(
        allow_undefined=True, template_loaders=[jinja2.DictLoader({"index.html": "Hello, {{ value }}!"})]
    )
    assert templates.render("index.html") == "Hello, !"


def test_undefined_off() -> None:
    templates = Templates(
        allow_undefined=False, template_loaders=[jinja2.DictLoader({"index.html": "Hello, {{ value }}!"})]
    )
    with pytest.raises(jinja2.exceptions.UndefinedError):
        assert templates.render("index.html") == "Hello, <world>!"


def test_autoescape_on() -> None:
    templates = Templates(auto_escape=True, template_loaders=[jinja2.DictLoader({"index.html": "Hello, {{ value }}!"})])
    assert templates.render("index.html", {"value": "<world>"}) == "Hello, &lt;world&gt;!"


def test_autoescape_off() -> None:
    templates = Templates(
        auto_escape=False, template_loaders=[jinja2.DictLoader({"index.html": "Hello, {{ value }}!"})]
    )
    assert templates.render("index.html", {"value": "<world>"}) == "Hello, <world>!"


def test_context_processors() -> None:
    def context_processor(request: Request) -> dict[str, typing.Any]:
        return {"name": "world"}

    templates = Templates(
        template_loaders=[jinja2.DictLoader({"index.html": "Hello, {{ name }}!"})],
        context_processors=[context_processor],
    )
    request = Request({"type": "http", "method": "GET", "url": "http://testserver/", "app": Kupala("key")})
    assert templates.render_to_response(request, "index.html", {"name": "world"}).body == b"Hello, world!"


def test_render() -> None:
    templates = Templates(jinja_env=jinja_env)
    assert templates.render("index.html", {"name": "world"}) == "Hello, world!"


def test_render_macro() -> None:
    templates = Templates(jinja_env=jinja_env)
    assert templates.render_macro("macro.html", "hello", {"name": "world"}) == "Hello, world!"


def test_render_block() -> None:
    templates = Templates(jinja_env=jinja_env)
    assert templates.render_block("block.html", "content", {"name": "world"}) == "Hello, world!"


def test_render_to_response() -> None:
    templates = Templates(jinja_env=jinja_env)
    request = Request({"type": "http", "method": "GET", "url": "http://testserver/", "app": Kupala("key")})
    assert templates.render_to_response(request, "index.html", {"name": "world"}).body == b"Hello, world!"


async def view(request: Request) -> Response:
    return PlainTextResponse("ok")


class TestStandardContextProcessor:
    def test_request(self) -> None:
        templates = Templates(jinja_env=jinja_env)
        request = Request(
            {"type": "http", "method": "GET", "url": "http://testserver/", "app": Kupala(secret_key="secret")}
        )
        assert b"starlette.requests.Request" in templates.render_to_response(request, "request.html").body

    def test_app(self) -> None:
        templates = Templates(jinja_env=jinja_env)
        request = Request(
            {"type": "http", "method": "GET", "url": "http://testserver/", "app": Kupala(secret_key="secret")}
        )
        assert b"kupala.applications.Kupala" in templates.render_to_response(request, "app.html").body

    def test_url_matches(self) -> None:
        templates = Templates(jinja_env=jinja_env)
        request = Request(
            {
                "type": "http",
                "method": "GET",
                "url": "http://testserver/",
                "headers": ((b"host", b"testserver"),),
                "path": "/test",
                "app": Kupala(
                    secret_key="secret",
                    routes=[
                        Route("/test", view),
                    ],
                ),
            }
        )
        assert b"True - False" in templates.render_to_response(request, "url_matches.html").body

    def test_route_matches(self) -> None:
        templates = Templates(jinja_env=jinja_env)
        request = Request(
            {
                "type": "http",
                "method": "GET",
                "url": "http://testserver/",
                "headers": ((b"host", b"testserver"),),
                "path": "/test",
                "app": Kupala(
                    secret_key="secret",
                    routes=[
                        Route("/test", view, name="test"),
                        Route("/no-test", view, name="no-test"),
                    ],
                ),
            }
        )
        assert b"True - False" in templates.render_to_response(request, "route_matches.html").body



import anyio
from starlette.responses import PlainTextResponse
from starlette.types import ASGIApp, Receive, Scope, Send


class RequestTimeoutMiddleware:
    def __init__(self, app: ASGIApp, timeout: float | None = None) -> None:
        """Limit the request processing time."""
        self.app = app
        self.timeout = timeout

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        try:
            timeout = self.timeout or scope["app"].request_timeout
            with anyio.fail_after(timeout):
                await self.app(scope, receive, send)
        except TimeoutError:
            response = PlainTextResponse("Gateway Timeout", status_code=504)
            await response(scope, receive, send)



import anyio
import pytest
from starlette.responses import PlainTextResponse
from starlette.testclient import TestClient
from starlette.types import Receive, Scope, Send

from kupala.asgi.request_timeout import RequestTimeoutMiddleware


async def app(scope: Scope, receive: Receive, send: Send) -> None:
    await anyio.sleep(0.2)
    await PlainTextResponse("")(scope, receive, send)


@pytest.mark.asyncio
async def test_timeout_middleware() -> None:
    client = TestClient(RequestTimeoutMiddleware(app, timeout=0.1))
    assert client.get("/").status_code == 504


@pytest.mark.asyncio
async def test_timeout_middleware_not_fails_withing_timespan() -> None:
    client = TestClient(RequestTimeoutMiddleware(app, timeout=1))
    assert client.get("/").status_code == 200



from __future__ import annotations

import abc
import asyncio
import base64
import binascii
import dataclasses
import datetime
import hashlib
import hmac
import typing

import blinker
import click
import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Mapped, mapped_column
from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
)
from starlette.authentication import (
    AuthenticationError as StarletteAuthenticationError,
)
from starlette.authentication import (
    BaseUser as StarletteBaseUser,
)
from starlette.middleware import Middleware as StarletteMiddleware
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.requests import HTTPConnection, Request
from starlette.responses import Response

from kupala.applications import Kupala
from kupala.database.extension import SQLAlchemy
from kupala.database.manager import on_commit
from kupala.database.models import Base, WithTimestamps
from kupala.dependencies import VariableResolver
from kupala.exceptions import HTTPException, NotAuthenticated
from kupala.passwords import PasswordHasher
from kupala.responses import JSONErrorResponse

user_authenticated = blinker.signal("user-authenticated", doc="User authenticated by an authenticator.")
user_registered = blinker.signal("user-registered", doc="User registered.")
user_logged_out = blinker.signal("user-logged-out", doc="User logged out.")
user_logged_in = blinker.signal("user-logged-in", doc="User interactively logged in.")
user_login_failed = blinker.signal("user-login-failed", doc="Interactive login failed.")
user_deactivated = blinker.signal("user-deactivated", doc="User deactivated.")
user_reactivated = blinker.signal("user-reactivated", doc="User reactivated.")
user_password_changed = blinker.signal("user-password-changed", doc="User password changed.")
user_password_reset = blinker.signal("user-password-reset", doc="User password reset.")
user_email_confirmed = blinker.signal("user-email-confirmed", doc="User email confirmed.")


class BaseUser(Base, StarletteBaseUser, WithTimestamps):
    """Base class for user objects."""

    __abstract__ = True
    __tablename__ = "users"
    __table_args__ = (sa.Index("users_email_udx", sa.func.lower(sa.literal_column("email")), unique=True),)

    email: Mapped[str] = mapped_column(doc="Email address.")
    password: Mapped[str] = mapped_column(doc="Password hash.")
    deactivated_at: Mapped[datetime.datetime | None] = mapped_column(doc="Deactivation timestamp.")
    email_verified_at: Mapped[datetime.datetime | None] = mapped_column(doc="Email verification timestamp.")

    @property
    def identity(self) -> str:
        return self.email

    @property
    def display_name(self) -> str:
        return self.email

    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def is_active(self) -> bool:
        return self.deactivated_at is None

    @property
    def is_email_confirmed(self) -> bool:
        return self.email_verified_at is not None

    def get_password_hash(self) -> str:
        return self.password

    def get_scopes(self) -> list[str]:
        return []

    @classmethod
    async def register(
        cls, dbsession: AsyncSession, *, email: str, password_hash: str, **attrs: typing.Any
    ) -> typing.Self:
        user = cls(email=email, password=password_hash, **attrs)
        dbsession.add(user)
        await dbsession.flush()

        def _after_commit() -> None:
            asyncio.create_task(user_registered.send_async(cls, user=user))

        on_commit(dbsession, _after_commit)
        return user

    def __str__(self) -> str:
        return self.display_name


type ByIDLoader = typing.Callable[[HTTPConnection, str], typing.Awaitable[BaseUser | None]]
type ByUserNameLoader = typing.Callable[[HTTPConnection, str], typing.Awaitable[BaseUser | None]]
type ByAPITokenLoader = typing.Callable[[HTTPConnection, str], typing.Awaitable[BaseUser | None]]

SESSION_IDENTITY_KEY = "kupala.identity"
SESSION_HASH_KEY = "kupala.identity_hash"


_U = typing.TypeVar("_U", bound=BaseUser, default=BaseUser)


class AuthenticationError(HTTPException, StarletteAuthenticationError):
    """Base class for authentication errors."""


class NotAuthenticatedError(AuthenticationError, NotAuthenticated):
    """User is not authenticated."""


@dataclasses.dataclass
class Identity(typing.Generic[_U]):
    id: str
    user: _U
    scopes: typing.Sequence[str]
    authenticator: type[Authenticator]


class DatabaseUserLoader(typing.Generic[_U]):
    def __init__(
        self,
        model_class: type[_U],
        where_builder: typing.Callable[[str], sa.ColumnElement[bool]],
    ) -> None:
        self.model_class = model_class
        self.where_builder = where_builder

    async def __call__(self, request: HTTPConnection, identity: str) -> _U | None:
        dbsession: AsyncSession = request.state.dbsession
        stmt = sa.select(self.model_class).where(self.where_builder(identity))
        result = await dbsession.scalars(stmt)
        return result.one_or_none()


def make_session_auth_hash(user: _U, secret_key: str) -> str:
    """Compute current user session auth hash."""
    key = hashlib.sha256(("kupala.auth." + secret_key).encode()).digest()
    return hmac.new(key, msg=user.get_password_hash().encode(), digestmod=hashlib.sha256).hexdigest()


def update_session_auth_hash(connection: HTTPConnection, user: _U, secret_key: str) -> None:
    """Update session auth hash.
    Call this function each time you change user's password.
    Otherwise, the session will be instantly invalidated."""
    connection.session[SESSION_HASH_KEY] = make_session_auth_hash(user, secret_key)


def validate_session_auth_hash(connection: HTTPConnection, session_auth_hash: str) -> bool:
    """Validate session auth hash."""
    return hmac.compare_digest(connection.session.get(SESSION_HASH_KEY, ""), session_auth_hash)


class Authenticator(abc.ABC):
    """Base class for authenticators."""

    @abc.abstractmethod
    async def authenticate(self, conn: HTTPConnection) -> Identity | None:
        raise NotImplementedError


class SessionAuthenticator(Authenticator):
    """Authenticates users based on session information.
    This authenticator looks for the user's ID in the session and loads the
    corresponding user object. It also validates the session authentication hash
    if the user object supports password hashing.
    """

    def __init__(self, user_loader: ByIDLoader) -> None:
        self.user_loader = user_loader

    async def authenticate(self, conn: HTTPConnection) -> Identity | None:
        user_id: str = conn.session.get(SESSION_IDENTITY_KEY, "")
        if not user_id:
            return None

        user = await self.user_loader(conn, user_id)
        if not user:
            return None

        # avoid authentication if session hash is invalid
        # this may happen when user changes password OR session is hijacked
        secret_key = conn.app.secret_key
        if conn.session.get(SESSION_HASH_KEY):
            if not validate_session_auth_hash(conn, make_session_auth_hash(user, secret_key)):
                return None

        scopes = user.get_scopes()

        return Identity(id=user.identity, user=user, scopes=scopes, authenticator=self.__class__)


class HTTPBasicAuthenticator(Authenticator):
    """
    A basic HTTP authentication authenticator.

    This authenticator implements the HTTP Basic Authentication scheme as specified in RFC 7617.
    It extracts credentials from the Authorization header, decodes them from base64, and validates
    the username/password against a provided user loader.

    If a session is available, the authentication hash will be updated in the session.
    """

    def __init__(self, user_loader: ByIDLoader) -> None:
        self.user_loader = user_loader

    async def authenticate(self, conn: HTTPConnection) -> Identity | None:
        auth = self.get_authorization_header(conn).split()

        if not auth or auth[0].lower() != "basic":
            return None

        if len(auth) == 1:
            raise AuthenticationError("Invalid basic header. No credentials provided.")

        if len(auth) > 2:
            raise AuthenticationError("Invalid basic header. No spaces allowed.")

        try:
            auth_decoded = base64.b64decode(auth[1]).decode("utf-8")
            identity, password = auth_decoded.split(":", 1)
        except (TypeError, ValueError, UnicodeDecodeError, binascii.Error):
            raise AuthenticationError("Invalid basic header. Credentials not correctly base64 encoded.")

        return await self.authenticate_credentials(conn, identity, password)

    def get_authorization_header(self, conn: HTTPConnection) -> str:
        return conn.headers.get("Authorization", "")

    async def authenticate_credentials(self, conn: HTTPConnection, identity: str, password: str) -> Identity | None:
        user = await self.user_loader(conn, identity)
        if not user:
            return None

        password_hasher = Authentication.of(conn.app).passwords
        if not await password_hasher.averify(password, user.get_password_hash()):
            return None

        if "session" in conn.scope:
            secret_key = conn.app.secret_key
            update_session_auth_hash(conn, user, secret_key)

        scopes = user.get_scopes()

        return Identity(id=user.identity, user=user, scopes=scopes, authenticator=self.__class__)


class ChainAuthenticator(Authenticator):
    """Authenticate user using multiple authenticators."""

    def __init__(self, *authenticators: Authenticator) -> None:
        self.authenticators = authenticators

    async def authenticate(self, conn: HTTPConnection) -> Identity | None:
        for authenticator in self.authenticators:
            if identity := await authenticator.authenticate(conn):
                return identity
        return None


class AuthenticatorBackend(AuthenticationBackend):
    """Integrates authenticators with Starlette's authentication middleware."""

    def __init__(self, authenticator: Authenticator) -> None:
        self.authenticator = authenticator

    async def authenticate(self, conn: HTTPConnection) -> tuple[AuthCredentials, BaseUser] | None:
        if identity := await self.authenticator.authenticate(conn):
            await user_authenticated.send_async(self.__class__, user=identity.user)
            return AuthCredentials(scopes=identity.scopes), identity.user
        return None


# TODO: OpenID connect authenticators: google, github, etc.
# TODO: run authorization as route middleware


class Authentication:
    def __init__(
        self,
        authenticator: Authenticator,
        password_hasher: PasswordHasher,
        user_model_class: type[_U],
    ) -> None:
        self.authenticator = authenticator
        self.passwords = password_hasher
        self.user_model_class = user_model_class

    def configure(self, app: Kupala) -> None:
        app.state[Authentication] = self
        app.commands.append(passwords_command)
        app.commands.append(users_command)
        app.dependencies.registry[PasswordHasher] = VariableResolver(self.passwords)
        app.asgi_middleware.append(
            StarletteMiddleware(
                AuthenticationMiddleware,
                backend=AuthenticatorBackend(self.authenticator),
            )
        )

    @classmethod
    def of(cls, app: Kupala) -> typing.Self:
        return typing.cast(typing.Self, app.state[Authentication])


def authentication_error_json_handler(request: Request, exc: Exception) -> Response:
    return JSONErrorResponse(message="Not authenticated.", status_code=401)


passwords_command = click.Group("passwords", help="Password management commands.")
users_command = click.Group("users", help="User management commands.")


@passwords_command.command("hash")
@click.argument("password")
@click.pass_obj
def hash_password_command(app: Kupala, password: str) -> None:
    """Hash a password."""
    passwords = Authentication.of(app).passwords
    click.echo(passwords.make(password))


@passwords_command.command("verify")
@click.argument("hashed_password")
@click.argument("plain_password")
@click.pass_obj
def verify_password_command(app: Kupala, hashed_password: str, plain_password: str) -> None:
    """Verify a password."""
    passwords = Authentication.of(app).passwords
    valid, new_hash = passwords.verify_and_migrate(plain_password, hashed_password)
    click.echo(
        "Valid: {valid}".format(valid=click.style("yes", fg="green") if valid else click.style("no", fg="yellow"))
    )
    if valid:
        click.echo(
            "Needs migration: {value}".format(
                value=click.style("yes", fg="yellow") if new_hash else click.style("no", fg="green")
            )
        )


@users_command.command("new")
@click.argument("email")
@click.argument("password")
@click.option("--dbname", default="default", help="Database name.")
@click.pass_obj
async def new_user_command(app: Kupala, email: str, password: str, dbname: str) -> None:
    """Create a new user."""
    passwords = Authentication.of(app).passwords
    user_model_class = Authentication.of(app).user_model_class
    database = SQLAlchemy.of(app).get_database(dbname)
    async with database.session() as dbsession, dbsession:
        user = await user_model_class.register(
            dbsession,
            email=email,
            password_hash=passwords.make(password),
        )
        await dbsession.commit()

    click.echo(
        "User {user} created, id={id}".format(
            user=click.style(user, fg="yellow"), id=click.style(user.identity, fg="green")
        )
    )


def _read_user_from_request(request: Request) -> BaseUser:
    user = request.user
    if not user.is_authenticated:
        raise NotAuthenticatedError("User is not authenticated.")
    return typing.cast(BaseUser, user)


type CurrentUser[_U] = typing.Annotated[_U, _read_user_from_request]



import functools
import os
import typing

import jinja2
import jinja2.ext
from starlette.requests import Request
from starlette.responses import Response
from starlette.templating import Jinja2Templates

from kupala import Kupala
from kupala.dependencies import VariableResolver
from kupala.routing import route_matches, url_matches

type ContextProcessor = typing.Callable[[Request], dict[str, typing.Any]]


def standard_processors(request: Request) -> dict[str, typing.Any]:
    return {
        "request": request,
        "app": request.app,
        "url_matches": functools.partial(url_matches, request),
        "route_matches": functools.partial(route_matches, request),
    }


class Templates(Jinja2Templates):
    def __init__(
        self,
        jinja_env: jinja2.Environment | None = None,
        *,
        debug: bool = False,
        auto_escape: bool = True,
        directories: typing.Sequence[str | os.PathLike[str]] = (),
        template_packages: typing.Sequence[str] = (),
        template_loaders: typing.Sequence[jinja2.BaseLoader] = (),
        context_processors: typing.Sequence[ContextProcessor] = (),
        extensions: typing.Sequence[str | type[jinja2.ext.Extension]] = (),
        globals: typing.Mapping[str, typing.Any] | None = None,
        filters: typing.Mapping[str, typing.Callable[[typing.Any], typing.Any]] | None = None,
        tests: typing.Mapping[str, typing.Callable[[typing.Any], bool]] | None = None,
        allow_undefined: bool = False,
    ) -> None:
        if not jinja_env:
            jinja_env = jinja2.Environment(
                auto_reload=debug,
                autoescape=auto_escape,
                extensions=extensions,
                undefined=jinja2.Undefined if allow_undefined else jinja2.StrictUndefined,
                loader=jinja2.ChoiceLoader(
                    [
                        *template_loaders,
                        jinja2.FileSystemLoader(directories),
                        *[jinja2.PackageLoader(package) for package in template_packages],
                        jinja2.PackageLoader("kupala"),
                    ]
                ),
            )
            jinja_env.globals.update(globals or {})
            jinja_env.filters.update(filters or {})
            jinja_env.tests.update(tests or {})

        context_processors = [standard_processors, *context_processors]

        super().__init__(
            env=jinja_env,
            context_processors=list(context_processors),
        )

    def render(self, name: str, context: dict[str, typing.Any] | None = None) -> str:
        """Render a template with the given context."""
        template = self.env.get_template(name)
        return template.render(context or {})

    def render_macro(
        self,
        name: str,
        macro: str,
        args: typing.Mapping[str, typing.Any] | None = None,
    ) -> str:
        """Render a macro from a template with the given arguments."""
        template: jinja2.Template = self.env.get_template(name)
        template_module = template.make_module({})
        callback = getattr(template_module, macro)
        return typing.cast(str, callback(**args or {}))

    def render_block(
        self,
        name: str,
        block: str,
        context: dict[str, typing.Any] | None = None,
    ) -> str:
        """Render a block from a template with the given context."""
        template = self.env.get_template(name)
        callback = template.blocks[block]
        template_context = template.new_context(context or {})
        return "".join(callback(template_context))

    def render_to_response(
        self,
        request: Request,
        name: str,
        context: dict[str, typing.Any] | None = None,
        status_code: int = 200,
        headers: typing.Mapping[str, str] | None = None,
        media_type: str | None = None,
    ) -> Response:
        """Render a template to a response."""
        return super().TemplateResponse(
            request,
            name,
            context or {},
            status_code=status_code,
            headers=headers,
            media_type=media_type,
        )

    @classmethod
    def of(cls, app: Kupala) -> typing.Self:
        return typing.cast(typing.Self, app.state[Templates])

    def configure(self, app: Kupala) -> None:
        app.state[Templates] = self
        app.state[jinja2.Environment] = self.env
        app.dependencies.registry[Templates] = VariableResolver(self)



import pytest
from starlette.requests import Request
from starlette.responses import Response
from starlette.testclient import TestClient
from starsessions import CookieStore, load_session, SessionNotLoaded

from kupala import Kupala
from kupala.routing import Route
from kupala.sessions import Sessions


async def set_value_view(request: Request) -> Response:
    request.session["test"] = "test"
    return Response("ok")


async def get_value_view(request: Request) -> Response:
    return Response(request.session["test"])


async def set_load_value_view(request: Request) -> Response:
    await load_session(request)
    request.session["test"] = "test"
    return Response("ok")


async def get_load_value_view(request: Request) -> Response:
    await load_session(request)
    return Response(request.session["test"])


def test_sessions() -> None:
    app = Kupala(
        secret_key="key!",
        routes=[
            Route("/set", set_value_view),
            Route("/get", get_value_view),
            Route("/set-load", set_load_value_view),
            Route("/get-load", get_load_value_view),
        ],
        extensions=[
            Sessions(
                store=CookieStore("key!"),
                cookie_https_only=False,
            )
        ],
    )
    with TestClient(app) as client:
        with pytest.raises(SessionNotLoaded):
            response = client.get("/set")
            assert response.status_code == 500

        response = client.get("/set-load")
        assert response.status_code == 200

        response = client.get("/get-load")
        assert response.status_code == 200
        assert response.text == "test"


def test_sessions_autoload() -> None:
    app = Kupala(
        secret_key="key!",
        routes=[
            Route("/set", set_value_view),
            Route("/get", get_value_view),
        ],
        extensions=[
            Sessions(
                autoload=True,
                store=CookieStore("key!"),
                cookie_https_only=False,
            )
        ],
    )
    with TestClient(app) as client:
        response = client.get("/set")
        assert response.status_code == 200

        response = client.get("/get")
        assert response.status_code == 200
        assert response.text == "test"


def test_sessions_cookie_name() -> None:
    app = Kupala(
        secret_key="key!",
        routes=[
            Route("/set", set_value_view),
        ],
        extensions=[Sessions(autoload=True, store=CookieStore("key!"), cookie_https_only=False, cookie_name="sid")],
    )
    with TestClient(app) as client:
        response = client.get("/set")
        assert response.status_code == 200
        assert "sid" in response.cookies


def test_sessions_cookie_same_site() -> None:
    app = Kupala(
        secret_key="key!",
        routes=[
            Route("/set", set_value_view),
        ],
        extensions=[
            Sessions(autoload=True, store=CookieStore("key!"), cookie_https_only=False, cookie_same_site="lax")
        ],
    )
    with TestClient(app) as client:
        response = client.get("/set")
        assert response.status_code == 200
        assert "lax" in response.headers["set-cookie"]


def test_sessions_cookie_domain() -> None:
    app = Kupala(
        secret_key="key!",
        routes=[
            Route("/set", set_value_view),
        ],
        extensions=[
            Sessions(autoload=True, store=CookieStore("key!"), cookie_https_only=False, cookie_domain="testserver")
        ],
    )
    with TestClient(app) as client:
        response = client.get("/set")
        assert response.status_code == 200
        assert response.cookies.get("session", domain=".testserver")


def test_sessions_cookie_path() -> None:
    app = Kupala(
        secret_key="key!",
        routes=[
            Route("/set", set_value_view),
        ],
        extensions=[Sessions(autoload=True, store=CookieStore("key!"), cookie_https_only=False, cookie_path="/secure")],
    )
    with TestClient(app) as client:
        response = client.get("/set")
        assert response.status_code == 200
        assert response.cookies.get("session", path="/secure")



from __future__ import annotations

import dataclasses
import typing

from starlette.exceptions import HTTPException as BaseHTTPException
from starlette import status


T = typing.TypeVar("T", default=str, bound=str | int)


@dataclasses.dataclass(frozen=True, slots=True)
class ErrorCode(typing.Generic[T]):
    code: T
    description: str = ""

    def __str__(self) -> str:
        return str(self.description)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str | int):
            return other == self.code

        if not isinstance(other, ErrorCode):
            raise NotImplementedError

        return bool(self.code == other.code)


class KupalaError(Exception):
    """Base class for all exceptions in the application.

    Message precedence:
    1. message argument
    2. error_code.description
    3. self.message
    """

    error_code: ErrorCode | None = None
    message: str | None = None

    def __init__(
        self,
        message: str | None = None,
        *,
        error_code: ErrorCode | None = None,
    ) -> None:
        error_code = error_code or self.error_code
        error_message = error_code.description if error_code else ""
        self.message = str(message or error_message or self.message)
        self.error_code = error_code
        super().__init__(self.message)


class HTTPException(BaseHTTPException):
    """Base class for all HTTP exceptions in the application."""

    http_code: int | None = None

    def __init__(
        self,
        detail: str | None = None,
        *,
        status_code: int | None = None,
        headers: typing.Mapping[str, str] | None = None,
        error_code: ErrorCode | None = None,
        errors: dict[str, list[str]] | None = None,
    ) -> None:
        status_code = status_code or self.http_code or status.HTTP_500_INTERNAL_SERVER_ERROR
        self.error_code = error_code
        self.errors = errors or {}
        super().__init__(
            status_code=status_code,
            detail=str(detail) if detail else None,
            headers=headers,
        )


class BadRequest(HTTPException):
    """The request cannot be fulfilled due to bad syntax."""

    http_code = 400


class NotAuthenticated(HTTPException):
    """User is not authenticated."""

    http_code = 401


class NotAuthorized(HTTPException):
    """User has no permissions to perform the action."""

    http_code = 403


class NotFound(HTTPException):
    """Page not found."""

    http_code = 404


class MethodNotAllowed(HTTPException):
    """HTTP method is not allowed for the requested URL."""

    http_code = 405


class NotAcceptable(HTTPException):
    """The server cannot produce a response matching the list of acceptable values."""

    http_code = 406


class RequestTimeout(HTTPException):
    """The server timed out waiting for the request."""

    http_code = 408


class Conflict(HTTPException):
    """A request conflict with the current state of the server."""

    http_code = 409


class UnsupportedMediaType(HTTPException):
    """The server cannot process the request because the payload is in an unsupported format."""

    http_code = 415


class UnprocessableEntity(HTTPException):
    """The server cannot process the request because it is semantically erroneous."""

    http_code = 422


class ValidationError(UnprocessableEntity):
    """Raised when validation fails. Same as UnprocessableEntity."""


class TooManyRequests(HTTPException):
    """The client has sent too many requests in a given amount of time."""

    http_code = 429



import datetime
import typing

from starlette.middleware import Middleware
from starsessions import JsonSerializer, Serializer, SessionAutoloadMiddleware, SessionMiddleware, SessionStore

from kupala import Kupala


class Sessions:
    def __init__(
        self,
        store: SessionStore,
        *,
        autoload: bool = False,
        autoload_patterns: list[str | typing.Pattern[str]] | None = None,
        rolling: bool = True,
        lifetime: datetime.timedelta = datetime.timedelta(days=14),
        cookie_name: str = "session",
        cookie_same_site: str = "strict",
        cookie_https_only: bool = True,
        cookie_domain: str | None = None,
        cookie_path: str = "/",
        serializer: Serializer | None = None,
    ) -> None:
        self.rolling = rolling
        self.store = store
        self.lifetime = lifetime
        self.cookie_name = cookie_name
        self.cookie_same_site = cookie_same_site
        self.cookie_https_only = cookie_https_only
        self.cookie_domain = cookie_domain
        self.cookie_path = cookie_path
        self.serializer = serializer or JsonSerializer()
        self.autoload = autoload
        self.autoload_patterns = autoload_patterns or []

    def configure(self, app: Kupala) -> None:
        app.asgi_middleware.append(
            Middleware(
                SessionMiddleware,
                store=self.store,
                lifetime=self.lifetime,
                rolling=self.rolling,
                cookie_name=self.cookie_name,
                cookie_same_site=self.cookie_same_site,
                cookie_https_only=self.cookie_https_only,
                cookie_domain=self.cookie_domain,
                cookie_path=self.cookie_path,
                serializer=self.serializer,
            ),
        )
        if self.autoload:
            app.asgi_middleware.append(Middleware(SessionAutoloadMiddleware, paths=self.autoload_patterns))



from kupala.database.dependencies import DbSession
from kupala.database.manager import Database, on_commit
from kupala.database.models import Base, WithTimestamps
from kupala.database.query import Query
from kupala.database.extension import SQLAlchemy

__all__ = [
    "Database",
    "Base",
    "WithTimestamps",
    "Query",
    "DbSession",
    "SQLAlchemy",
    "on_commit",
]



import contextlib
import typing

from sqlalchemy.ext.asyncio import AsyncSession
from starlette.middleware import Middleware

from kupala.applications import Kupala
from kupala.database.manager import Database
from kupala.database.middleware import DbSessionMiddleware
from kupala.dependencies import RequestResolver


class SQLAlchemy:
    def __init__(self, databases: dict[str, Database]) -> None:
        self._databases = databases

    @contextlib.asynccontextmanager
    async def initialize(self, _: Kupala) -> typing.AsyncGenerator[None, None]:
        async with contextlib.AsyncExitStack() as stack:
            for database in self._databases.values():
                await stack.enter_async_context(database)
            yield

    def configure(self, app: Kupala) -> None:
        app.initializers.append(self.initialize)
        app.state[SQLAlchemy] = self
        app.dependencies.registry[AsyncSession] = RequestResolver(lambda r: r.state.dbsession)
        app.asgi_middleware.insert(0, Middleware(DbSessionMiddleware, self._databases))

    @classmethod
    def of(cls, app: Kupala) -> typing.Self:
        return typing.cast(typing.Self, app.state[SQLAlchemy])

    def get_database(self, name: str) -> Database:
        return self._databases[name]

    @classmethod
    def new_session(cls, name: str) -> typing.AsyncContextManager[AsyncSession]:
        """Create a new database session."""
        return cls.of(Kupala.current()).get_database(name).session()
